from classicML.api.plots import _plt as plt
from classicML.api.plots.utils import _set_history_axis_and_background
from classicML.api.plots.utils import _history_plot_config


def plot_history(history):
    """可视化历史记录.

    Arguments:
        history: classicML.backend.callbacks.History 实例.
    """
    _, ax = plt.subplots()
    _set_history_axis_and_background(ax)

    # 绘制损失曲线
    ax.plot(history.loss, label=history.loss_name, c='lightcoral')
    # 绘制评估曲线
    ax.plot(history.metric, label=history.metric_name, c='c')

    _history_plot_config()
    plt.show()



"""classicML中的benchmarks用于评估和测试模型的性能和开销"""
from classicML.benchmarks.wrapper_utils import average_timer
from classicML.benchmarks.wrapper_utils import memory_monitor
from classicML.benchmarks.wrapper_utils import timer



from classicML.api.plots import _plt as plt


def _set_bayes_axis(axes):
    """设置贝叶斯画布的坐标轴."""
    axes.spines['top'].set_color('none')
    axes.spines['right'].set_color('none')


def _set_lda_axis(axes):
    """设置线性判别分析画布的坐标轴."""
    axes.spines['top'].set_color('none')
    axes.spines['right'].set_color('none')


def _set_history_axis_and_background(axes):
    """设置历史记录画布的坐标轴."""
    axes.spines['left'].set_color('none')
    axes.spines['right'].set_color('none')
    axes.spines['top'].set_color('none')
    axes.spines['bottom'].set_color('none')

    axes.patch.set_facecolor('gray')
    axes.patch.set_alpha(0.1)  # 设置透明度
    axes.grid(axis='y', linestyle='dotted')


def _set_svc_axis(axes):
    """设置支持向量分类器画布的坐标轴.
    具体实现参看_set_history_axis_and_background
    """
    _set_history_axis_and_background(axes)


def _set_tree_axis(axes):
    """设置支持树画布的坐标轴."""
    axes.spines['left'].set_color('none')
    axes.spines['right'].set_color('none')
    axes.spines['top'].set_color('none')
    axes.spines['bottom'].set_color('none')

    axes.set_xticks([])
    axes.set_yticks([])


def _bayes_plot_config(x_label, y_label):
    """设置贝叶斯画布的其他配置项"""
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.legend()


def _history_plot_config():
    """设置历史记录画布的其他配置项.
    """
    plt.xlabel('epochs')
    plt.legend()


def _lda_plot_config(x_label, y_label):
    """设置线性判别分析画布的其他配置项"""
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.axis([0, 1, 0, 1])
    plt.legend()


def _logistic_regression_plot_config(x_label, y_label):
    """设置逻辑回归画布的其他配置项"""
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.axis([0, 1, 0, 1])
    plt.legend()


def _svc_plot_config(CS, kernel_name, C, x_label, y_label):
    """设置支持向量分类器的其他配置项."""
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.clabel(CS, fmt={CS.levels[0]: '决策边界'})
    plt.title('kernel={}, C={}'.format(kernel_name, C))
    plt.legend()


def _add_node(parent_node, child_node, node_name, node_type, axes):
    """添加结点信息.

     Arguments:
        parent_node: tuple,
            父结点的坐标.
        child_node: tuple,
            自身结点的坐标.
        node_name: str,
            当前结点的名称文本内容.
        node_type: dict,
            结点的类型信息(不同的结点使用不同的图例).
        axes: matplotlib.axes._subplots.AxesSubplot,
            matplotlib的绘图工具.
    """
    axes.annotate(node_name,
                  xy=[parent_node[0], parent_node[1] - 0.05],
                  xytext=child_node,
                  xycoords='axes fraction',
                  textcoords='axes fraction',
                  arrowprops=dict(arrowstyle='<-'),
                  verticalalignment='center',
                  horizontalalignment='center',
                  size=15,
                  bbox=node_type)


def _add_text(parent_node, child_node, text, axes):
    """添加文本信息.

    Arguments:
        parent_node: tuple,
            父结点的坐标.
        child_node: tuple,
            自身结点的坐标.
        text: str,
            文本内容.
        axes: matplotlib.axes._subplots.AxesSubplot,
            matplotlib的绘图工具.
    """
    x = (parent_node[0] - child_node[0]) / 2 + child_node[0]
    y = (parent_node[1] - child_node[1]) / 2 + child_node[1]

    axes.text(x, y, text, fontdict=dict(size=9))


def _tree_plot_config(criterion, pruning):
    """设置树的其他配置项."""
    if pruning:
        name = pruning.name
    else:
        name = None
    plt.title('criterion={}, pruning={}'.format(criterion, name),
              y=-0.1)



from classicML.api.plots import _plt as plt
from classicML.api.plots.utils import _add_node
from classicML.api.plots.utils import _add_text
from classicML.api.plots.utils import _set_tree_axis
from classicML.api.plots.utils import _tree_plot_config

NUM_OF_LEAVES = 0
HEIGHT = -1
X = 0.5
Y = 1


def _plot_decision_tree(decision_tree, parent_node, node_text, axes):
    """绘制决策树.

     Arguments:
        decision_tree: classicML.tree._TreeNode,
            决策树实例.
        parent_node: tuple,
            父结点的坐标.
        node_text: str,
            当前结点内的文本内容.
        axes: matplotlib.axes._subplots.AxesSubplot,
            matplotlib的绘图工具.
    """
    global NUM_OF_LEAVES
    global HEIGHT
    global X
    global Y

    child_node = (X + (1 + decision_tree.num_of_leaves) / (2 * NUM_OF_LEAVES), Y)
    _add_text(parent_node, child_node, node_text, axes)

    if HEIGHT == 0:
        # 决策树桩(相当于叶结点).
        _add_node(parent_node=parent_node,
                  child_node=child_node,
                  node_name=decision_tree.category,
                  node_type=dict(boxstyle='circle, pad=0.4', fc='orange'),
                  axes=axes)

        return
    else:
        # 决策的分支结点.
        _add_node(parent_node=parent_node,
                  child_node=child_node,
                  node_name=decision_tree.feature_name,
                  node_type=dict(boxstyle='round, pad=0.45', fc='lightcoral'),
                  axes=axes)

        Y -= 1 / HEIGHT

        for key in decision_tree.subtree.keys():
            if decision_tree.subtree[key].leaf:
                # 叶结点.
                X += 1 / NUM_OF_LEAVES

                _add_text(child_node, (X, Y), str(key), axes)
                _add_node(parent_node=child_node,
                          child_node=(X, Y),
                          node_name=decision_tree.subtree[key].category,
                          node_type=dict(boxstyle='circle, pad=0.4', fc='orange'),
                          axes=axes)
            else:
                _plot_decision_tree(decision_tree.subtree[key], child_node, str(key), axes)

        Y += 1 / HEIGHT


def plot_tree(tree):
    """绘制树的示意图.

    Arguments:
        tree: classicML.models.Tree, 树实例.
    """
    # 使用全局变量, 减少函数的传参并且使得递归函数更易理解.
    global NUM_OF_LEAVES
    global HEIGHT
    global X
    global Y

    decision_tree = tree.tree

    NUM_OF_LEAVES = decision_tree.num_of_leaves
    HEIGHT = decision_tree.height
    X = -0.5 / NUM_OF_LEAVES
    Y = 1

    _, ax = plt.subplots()
    _set_tree_axis(ax)

    # 绘制决策树.
    _plot_decision_tree(decision_tree, (0.5, 1), '', ax)

    _tree_plot_config(tree.criterion, tree.pruner)

    # 自动调整子图.
    plt.tight_layout()
    plt.show()



from pickle import loads, dumps

import numpy as np
import pandas as pd

from classicML import CLASSICML_LOGGER
from classicML.api.models import BaseModel
from classicML.backend import get_conditional_probability
from classicML.backend import get_dependent_prior_probability
from classicML.backend import get_probability_density
from classicML.backend import type_of_target
from classicML.backend import io


class OneDependentEstimator(BaseModel):
    """独依赖估计器的基类.

    Attributes:
        attribute_name: list of name, default=None,
            属性的名称.
        is_trained: bool, default=False,
            模型训练后将被标记为True.
        is_loaded: bool, default=False,
            如果模型加载了权重将被标记为True.

    Raises:
        NotImplementedError: compile, fit, predict方法需要用户实现.
    """
    def __init__(self, attribute_name=None):
        """初始化独依赖估计器.

        Arguments:
            attribute_name: list of name, default=None,
                属性的名称.
        """
        super(OneDependentEstimator, self).__init__()
        self.attribute_name = attribute_name

        self.is_trained = False
        self.is_loaded = False

    def compile(self, *args, **kwargs):
        """编译独依赖估计器.
        """
        raise NotImplementedError

    def fit(self, x, y, **kwargs):
        """训练独依赖估计器.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like, 特征数据.
            y: numpy.ndarray or pandas.DataFrame, array-like, 标签.
        """
        raise NotImplementedError

    def predict(self, x, **kwargs):
        """使用独依赖估计器进行预测.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like, 特征数据.
        """
        raise NotImplementedError

    def load_weights(self, filepath):
        """加载模型参数.

        Arguments:
            filepath: str, 权重文件加载的路径.

        Raises:
            KeyError: 模型权重加载失败.

        Notes:
            模型将不会加载关于优化器的超参数.
        """
        raise NotImplementedError

    def save_weights(self, filepath):
        """将模型权重保存为一个HDF5文件.

        Arguments:
            filepath: str, 权重文件保存的路径.

        Raises:
            TypeError: 模型权重保存失败.

        Notes:
            模型将不会保存关于优化器的超参数.
        """
        raise NotImplementedError


class SuperParentOneDependentEstimator(OneDependentEstimator):
    """超父独依赖估计器.

    Attributes:
        attribute_name: list of name, default=None,
            属性的名称.
        super_parent_name: str, default=None,
            超父的名称.
        super_parent_index: int, default=None,
            超父的索引值.
        _list_of_p_c: list,
            临时保存中间的概率依赖数据.
        smoothing: bool, default=None,
            是否使用平滑, 这里的实现是拉普拉斯修正.
    """
    def __init__(self, attribute_name=None):
        """初始化超父独依赖估计器.

        Arguments:
            attribute_name: list of name, default=None,
                属性的名称.
        """
        super(SuperParentOneDependentEstimator, self).__init__(attribute_name=attribute_name)

        self.super_parent_name = None
        self.super_parent_index = None
        self.smoothing = None

        self._list_of_p_c = list()

    def compile(self, super_parent_name, smoothing=True):
        """编译超父独依赖估计器.

        Arguments:
            super_parent_name: str, default=None,
                超父的名称.
            smoothing: bool, default=True,
                是否使用平滑, 这里的实现是拉普拉斯修正.
        """
        self.super_parent_name = super_parent_name
        self.smoothing = smoothing

    def fit(self, x, y, **kwargs):
        """训练超父独依赖估计器.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like, 特征数据.
            y: numpy.ndarray or pandas.DataFrame, array-like, 标签.

        Returns:
            SuperParentOneDependentEstimator实例.
        """
        if isinstance(x, np.ndarray) and self.attribute_name is None:
            CLASSICML_LOGGER.warn("属性名称缺失, 请使用pandas.DataFrame; 或检查 self.attributes_name")

        # 为特征数据添加属性信息.
        x = pd.DataFrame(x, columns=self.attribute_name)
        x.reset_index(drop=True, inplace=True)
        y = pd.Series(y)
        y.reset_index(drop=True, inplace=True)

        for index, feature_name in enumerate(x.columns):
            if self.super_parent_name == feature_name:
                self.super_parent_index = index

        for category in np.unique(y):
            unique_values_xi = x.iloc[:, self.super_parent_index].unique()
            for value in unique_values_xi:
                # 初始化概率字典.
                p_c = dict()

                # 获取有依赖的类先验概率P(c, xi).
                c_xi = (x.values[:, self.super_parent_index] == value) & (y == category)
                c_xi = x.values[c_xi, :]
                p_c_xi = get_dependent_prior_probability(len(c_xi),
                                                         len(x.values),
                                                         len(unique_values_xi),
                                                         self.smoothing)
                p_c.update({'p_c_xi': p_c_xi})

                # 获取有依赖的类条件概率P(xj|c, xi)或概率密度p(xj|c, xi)所需的信息.
                for attribute in range(x.shape[1]):
                    xj = x.iloc[:, attribute]
                    continuous = type_of_target(xj.values) == 'continuous'

                    if continuous:
                        # 连续值概率密度函数信息.
                        if len(c_xi) <= 2:
                            # 样本数量过少的时候, 使用全局的均值和方差.
                            mean = np.mean(x.values[y == category, attribute])
                            var = np.var(x.values[y == category, attribute])
                        else:
                            mean = np.mean(c_xi[:, attribute])
                            var = np.var(c_xi[:, attribute])
                        p_c.update({x.columns[attribute]: {
                                    'continuous': continuous,
                                    'values': [mean, var]}})
                    else:
                        # 离散值条件概率信息.
                        unique_value = xj.unique()
                        num_of_unique_value = len(unique_value)
                        value_count = pd.DataFrame(np.zeros((1, num_of_unique_value)), columns=unique_value)

                        for key in pd.value_counts(c_xi[:, attribute]).keys():
                            value_count[key] += pd.value_counts(c_xi[:, attribute])[key]

                        # 统计不同属性值的样本总数.
                        D_c_xi = dict()
                        for name in value_count:
                            D_c_xi.update({name: float(value_count[name].values)})

                        p_c.update({x.columns[attribute]: {
                                    'continuous': continuous,
                                    'values': [D_c_xi, c_xi.shape[0], num_of_unique_value],
                                    'smoothing': self.smoothing}})

                self._list_of_p_c.append({'category': category, 'attribute': value, 'p_c': p_c})

        self.is_trained = True

        return self

    def predict(self, x, probability=False):
        """使用超父独依赖估计器进行预测.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like, 特征数据.
            probability: bool, default=False,
                是否使用归一化的概率形式.

        Returns:
            SuperParentOneDependentEstimator的预测结果,
            不使用概率形式将返回0或1的标签数组, 使用将返回反正例概率的数组.

        Raises:
            ValueError: 模型没有训练的错误.
        """
        if self.is_trained is False and self.is_loaded is False:
            CLASSICML_LOGGER.error('模型没有训练')
            raise ValueError('你必须先进行训练')

        # 为特征数据添加属性信息.
        x = pd.DataFrame(x, columns=self.attribute_name)
        x.reset_index(drop=True, inplace=True)

        y_pred = list()

        if len(x.shape) == 1:
            p_0, p_1 = self._predict(x)
            if probability:
                y_pred.append([p_0 / (p_0 + p_1), p_1 / (p_0 + p_1)])
            else:
                if p_0 > p_1:
                    y_pred.append(0)
                else:
                    y_pred.append(1)
        else:
            for i in range(x.shape[0]):
                x_test = x.iloc[i, :]
                p_0, p_1 = self._predict(x_test)
                if probability:
                    y_pred.append([p_0 / (p_0 + p_1), p_1 / (p_0 + p_1)])
                else:
                    if p_0 > p_1:
                        y_pred.append(0)
                    else:
                        y_pred.append(1)

        return y_pred

    def load_weights(self, filepath):
        """加载模型参数.

        Arguments:
            filepath: str, 权重文件加载的路径.

        Raises:
            KeyError: 模型权重加载失败.

        Notes:
            模型将不会加载关于优化器的超参数.
        """
        # 初始化权重文件.
        parameters_gp = io.initialize_weights_file(filepath=filepath,
                                                   mode='r',
                                                   model_name='SuperParentOneDependentEstimator')
        # 加载模型参数.
        try:
            compile_ds = parameters_gp['compile']
            weights_ds = parameters_gp['weights']

            self.super_parent_name = compile_ds.attrs['super_parent_name']
            self.smoothing = compile_ds.attrs['smoothing']
            self._list_of_p_c = loads(weights_ds.attrs['_list_of_p_c'].tobytes())

            # 标记加载完成
            self.is_loaded = True
        except KeyError:
            CLASSICML_LOGGER.error('模型权重加载失败, 请检查文件是否损坏')
            raise KeyError('模型权重加载失败')

    def save_weights(self, filepath):
        """将模型权重保存为一个HDF5文件.

        Arguments:
            filepath: str, 权重文件保存的路径.

        Raises:
            TypeError: 模型权重保存失败.

        Notes:
            模型将不会保存关于优化器的超参数.
        """
        # 初始化权重文件.
        parameters_gp = io.initialize_weights_file(filepath=filepath,
                                                   mode='w',
                                                   model_name='SuperParentOneDependentEstimator')
        # 保存模型参数.
        try:
            compile_ds = parameters_gp['compile']
            weights_ds = parameters_gp['weights']

            compile_ds.attrs['super_parent_name'] = self.super_parent_name
            compile_ds.attrs['smoothing'] = self.smoothing
            weights_ds.attrs['_list_of_p_c'] = np.void(dumps(self._list_of_p_c))
        except TypeError:
            CLASSICML_LOGGER.error('模型权重保存失败, 请检查文件是否损坏')
            raise TypeError('模型权重保存失败')

    def _predict(self, x, attribute_list=None, super_parent_index=None):
        """通过平均独依赖估计器预测单个样本.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like,
                特征数据.
            attribute_list: list, default=None,
                临时保存中间的概率依赖数据(包含所有的属性, 仅使用AODE时有意义).
            super_parent_index: int, default=None,
                超父的索引值.

        Returns:
            返回预测的结果.
        """
        y_pred = [0.0, 0.0]

        if attribute_list is None and super_parent_index is None:
            for i in self._list_of_p_c:
                self._calculate_posterior_probability(x, i, y_pred)
        else:
            # TODO(Steve R. Sun, tag:performance): 这里是为了满足AODE调用的便利.
            for i in attribute_list[super_parent_index]:
                if i['attribute'] == x[super_parent_index]:
                    self._calculate_posterior_probability(x, i, y_pred)

        return y_pred

    @staticmethod
    def _calculate_posterior_probability(x, i, y_pred):
        """计算后验概率.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like,
                特征数据.
            i: int, 样本的索引.
            y_pred: list, 后验概率列表.
        """
        _p_c = i['p_c']

        if i['category'] == 0:
            for index, probability in enumerate(_p_c):
                if probability == 'p_c_xi':
                    # 先添加P(c, xi)
                    y_pred[0] += np.log(_p_c[probability])
                else:
                    # 添加P(xj|c, xi)
                    continuous = _p_c[probability]['continuous']
                    # 分别处理连续值和离散值.
                    if continuous:
                        mean, var = _p_c[probability]['values']
                        probability_density = get_probability_density(x[index - 1], mean, var)
                        y_pred[0] += np.log(probability_density)  # 存放数据中多存放一个p_c_xi导致和x的索引无法对齐.
                    else:
                        D_c_xi_xj, D_c_xi, num_of_unique_value = _p_c[probability]['values']
                        y_pred[0] += np.log(get_conditional_probability(D_c_xi_xj[x[index - 1]],
                                                                        D_c_xi,
                                                                        num_of_unique_value,
                                                                        _p_c[probability]['smoothing']))
        elif i['category'] == 1:
            for index, probability in enumerate(_p_c):
                if probability == 'p_c_xi':
                    y_pred[1] += np.log(_p_c[probability])
                else:
                    continuous = _p_c[probability]['continuous']
                    if continuous:
                        mean, var = _p_c[probability]['values']
                        probability_density = get_probability_density(x[index - 1], mean, var)
                        y_pred[1] += np.log(probability_density)
                    else:
                        D_c_xi_xj, D_c_xi, num_of_unique_value = _p_c[probability]['values']
                        y_pred[1] += np.log(get_conditional_probability(D_c_xi_xj[x[index - 1]],
                                                                        D_c_xi,
                                                                        num_of_unique_value,
                                                                        _p_c[probability]['smoothing']))


class AveragedOneDependentEstimator(SuperParentOneDependentEstimator):
    """平均独依赖估计器.

    Attributes:
        attribute_name: list of name, default=None,
            属性的名称.
        super_parent_name: str, default=None,
            超父的名称.
        smoothing: bool, default=None,
            是否使用平滑, 这里的实现是拉普拉斯修正.
        m: int, default=0,
            阈值常数, 样本小于此值的属性将不会被作为超父类.
        _attribute_list: list,
            临时保存中间的概率依赖数据(包含所有的属性).
    """
    def __init__(self, attribute_name=None):
        """初始化平均独依赖估计器.

        Arguments:
            attribute_name: list of name, default=None,
                属性的名称.
        """
        super(AveragedOneDependentEstimator, self).__init__(attribute_name=attribute_name)

        self.smoothing = None
        self.m = 0
        self._attribute_list = list()

    def compile(self, smoothing=True, m=0, **kwargs):
        """编译平均独依赖估计器.

        Arguments:
            smoothing: bool, default=True,
                是否使用平滑, 这里的实现是拉普拉斯修正.
            m: int, default=0,
                阈值常数, 样本小于此值的属性将不会被作为超父类.
        """
        self.smoothing = smoothing
        self.m = m

    def fit(self, x, y, **kwargs):
        """训练平均独依赖估计器.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like, 特征数据.
            y: numpy.ndarray or pandas.DataFrame, array-like, 标签.

        Returns:
            AverageOneDependentEstimator实例.
        """
        if isinstance(x, np.ndarray) and self.attribute_name is None:
            CLASSICML_LOGGER.warn("属性名称缺失, 请使用pandas.DataFrame; 或检查 self.attributes_name")

        # TODO(Steve R. Sun, tag:code): 暂时没有找到合理的断点续训的理论支持.
        self._attribute_list = list()

        # 为特征数据添加属性信息.
        x = pd.DataFrame(x, columns=self.attribute_name)
        x.reset_index(drop=True, inplace=True)
        y = pd.Series(y)
        y.reset_index(drop=True, inplace=True)

        number_of_samples, number_of_attributes = x.shape

        # 获取离散属性的全部取值.
        discrete_unique_values = dict()
        for attribute in range(number_of_attributes):
            xi = x.iloc[:, attribute]
            if (type_of_target(xi.values) != 'continuous') and (pd.value_counts(xi).values > self.m).all():
                discrete_unique_values.update({x.columns[attribute]: xi.unique()})

        # 每个属性作为超父类构建SPODE.
        for index, key in enumerate(discrete_unique_values.keys()):
            self.super_parent_name = key
            super(AveragedOneDependentEstimator, self).fit(x, y)
            current_attribute_list = self._list_of_p_c

            self._attribute_list.append(current_attribute_list)

        self.is_trained = True

        return self

    def predict(self, x, **kwargs):
        """使用平均独依赖估计器进行预测.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like, 特征数据.

        Returns:
            AverageOneDependentEstimator的预测结果.

        Raises:
            ValueError: 模型没有训练的错误.
        """
        if self.is_trained is False and self.is_loaded is False:
            CLASSICML_LOGGER.error('模型没有训练')
            raise ValueError('你必须先进行训练')

        # 为特征数据添加属性信息.
        x = pd.DataFrame(x, columns=self.attribute_name)
        x.reset_index(drop=True, inplace=True)

        y_pred = list()
        if len(x.shape) == 1:
            y_pred.append(self._predict(x))
        else:
            for i in range(x.shape[0]):
                x_test = x.iloc[i, :]
                y_pred.append(self._predict(x_test))

        return y_pred

    def load_weights(self, filepath):
        """加载模型参数.

        Arguments:
            filepath: str, 权重文件加载的路径.

        Raises:
            KeyError: 模型权重加载失败.

        Notes:
            模型将不会加载关于优化器的超参数.
        """
        # 初始化权重文件.
        parameters_gp = io.initialize_weights_file(filepath=filepath,
                                                   mode='r',
                                                   model_name='AveragedOneDependentEstimator')
        # 加载模型参数.
        try:
            compile_ds = parameters_gp['compile']
            weights_ds = parameters_gp['weights']

            self.smoothing = compile_ds.attrs['smoothing']
            self.m = compile_ds.attrs['m']
            self._attribute_list = loads(weights_ds.attrs['_attribute_list'].tobytes())

            # 标记加载完成
            self.is_loaded = True
        except KeyError:
            CLASSICML_LOGGER.error('模型权重加载失败, 请检查文件是否损坏')
            raise KeyError('模型权重加载失败')

    def save_weights(self, filepath):
        """将模型权重保存为一个HDF5文件.

        Arguments:
            filepath: str, 权重文件保存的路径.

        Raises:
            TypeError: 模型权重保存失败.

        Notes:
            模型将不会保存关于优化器的超参数.
        """
        # 初始化权重文件.
        parameters_gp = io.initialize_weights_file(filepath=filepath,
                                                   mode='w',
                                                   model_name='AveragedOneDependentEstimator')
        # 保存模型参数.
        try:
            compile_ds = parameters_gp['compile']
            weights_ds = parameters_gp['weights']

            compile_ds.attrs['smoothing'] = self.smoothing
            compile_ds.attrs['m'] = self.m
            weights_ds.attrs['_attribute_list'] = np.void(dumps(self._attribute_list))
        except TypeError:
            CLASSICML_LOGGER.error('模型权重保存失败, 请检查文件是否损坏')
            raise TypeError('模型权重保存失败')

    def _predict(self, x, **kwargs):
        """
        通过平均独依赖估计器预测单个样本.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like, 特征数据.

        Returns:
            返回预测的结果.
        """
        avg_result = {'0': 0.0, '1': 0.0}
        for i in range(len(self._attribute_list)):
            _temp_y_pred = super(AveragedOneDependentEstimator, self)._predict(x, self._attribute_list, i)
            avg_result['0'] += _temp_y_pred[0]
            avg_result['1'] += _temp_y_pred[1]

        if avg_result['0'] > avg_result['1']:
            return 0
        else:
            return 1



import os

from classicML import CLASSICML_LOGGER

# python和cc多后端模块
if os.environ['CLASSICML_ENGINE'] == 'CC':
    from classicML.backend.cc import activations
    from classicML.backend.cc import callbacks
    from classicML.backend.python import data
    from classicML.backend.cc import initializers
    from classicML.backend.cc import kernels
    from classicML.backend.cc import losses
    from classicML.backend.cc import metrics
    from classicML.backend.python import optimizers
    from classicML.backend.python import tree

    from classicML.backend.cc.activations import __version__ as activations__version__
    from classicML.backend.cc.callbacks import __version__ as callbacks__version__
    from classicML.backend.cc.initializers import __version__ as initializers__version__
    from classicML.backend.cc.kernels import __version__ as kernels__version__
    from classicML.backend.cc.losses import __version__ as losses__version__
    from classicML.backend.cc.metrics import __version__ as metrics__version__

    CLASSICML_LOGGER.info('后端版本是: {}'.format(activations__version__))
    CLASSICML_LOGGER.info('后端版本是: {}'.format(callbacks__version__))
    CLASSICML_LOGGER.info('后端版本是: {}'.format(initializers__version__))
    CLASSICML_LOGGER.info('后端版本是: {}'.format(kernels__version__))
    CLASSICML_LOGGER.info('后端版本是: {}'.format(losses__version__))
    CLASSICML_LOGGER.info('后端版本是: {}'.format(metrics__version__))
else:
    from classicML.backend.python import activations
    from classicML.backend.python import callbacks
    from classicML.backend.python import data
    from classicML.backend.python import initializers
    from classicML.backend.python import kernels
    from classicML.backend.python import losses
    from classicML.backend.python import metrics
    from classicML.backend.python import optimizers
    from classicML.backend.python import tree

# ops模块
if os.environ['CLASSICML_ENGINE'] == 'CC':
    from classicML.backend.cc.ops import cc_calculate_error
    from classicML.backend.cc.ops import cc_clip_alpha
    from classicML.backend.cc.ops import cc_get_conditional_probability
    from classicML.backend.cc.ops import cc_get_dependent_prior_probability
    from classicML.backend.cc.ops import cc_get_prior_probability
    from classicML.backend.cc.ops import cc_get_probability_density
    from classicML.backend.cc.ops import cc_get_w as cc_get_w_v1  # 正式版将移除.
    from classicML.backend.cc.ops import cc_get_w_v2
    from classicML.backend.cc.ops import cc_get_w_v2 as cc_get_w
    from classicML.backend.cc.ops import cc_get_within_class_scatter_matrix
    from classicML.backend.cc.ops import cc_select_second_alpha
    from classicML.backend.cc.ops import cc_type_of_target

    from classicML.backend.cc.ops import cc_calculate_error as calculate_error
    from classicML.backend.cc.ops import cc_clip_alpha as clip_alpha
    from classicML.backend.cc.ops import cc_get_conditional_probability as get_conditional_probability
    from classicML.backend.cc.ops import cc_get_dependent_prior_probability as get_dependent_prior_probability
    from classicML.backend.cc.ops import cc_get_prior_probability as get_prior_probability
    from classicML.backend.cc.ops import cc_get_probability_density as get_probability_density
    from classicML.backend.cc.ops import cc_get_w_v2 as get_w
    from classicML.backend.cc.ops import cc_get_within_class_scatter_matrix as get_within_class_scatter_matrix
    from classicML.backend.cc.ops import cc_select_second_alpha as select_second_alpha
    from classicML.backend.cc.ops import cc_type_of_target as type_of_target

    from classicML.backend.cc.ops import __version__ as ops__version__
    CLASSICML_LOGGER.info('后端版本是: {}'.format(ops__version__))
else:
    from classicML.backend.python.ops import calculate_error
    from classicML.backend.python.ops import clip_alpha
    from classicML.backend.python.ops import get_conditional_probability
    from classicML.backend.python.ops import get_dependent_prior_probability
    from classicML.backend.python.ops import get_prior_probability
    from classicML.backend.python.ops import get_probability_density
    from classicML.backend.python.ops import get_w as get_w_v1  # 正式版将移除.
    from classicML.backend.python.ops import get_w_v2
    from classicML.backend.python.ops import get_w_v2 as get_w
    from classicML.backend.python.ops import get_within_class_scatter_matrix
    from classicML.backend.python.ops import select_second_alpha
    from classicML.backend.python.ops import type_of_target

from classicML.backend import io
from classicML.backend.training import get_initializer
from classicML.backend.training import get_kernel
from classicML.backend.training import get_loss
from classicML.backend.training import get_metric
from classicML.backend.training import get_optimizer
from classicML.backend.training import get_pruner



"""classicML的激活函数."""
import numpy as np


class Activation(object):
    """激活函数基类.

    Attributes:
        name: str, default='activation',
            激活函数名称.

    Raises:
       NotImplementedError: __call__, diff方法需要用户实现.
    """
    def __init__(self, name='activation'):
        self.name = name

    def __call__(self, z):
        """函数实现.

        Arguments:
            z: numpy.ndarray, 输入张量.
        """
        raise NotImplementedError

    def diff(self, output, a, *args, **kwargs):
        """函数的导数(微分).

        Arguments:
            output: numpy.ndarray, 输出张量.
            a: numpy.ndarray, 输入张量.
        """
        raise NotImplementedError


class Relu(Activation):
    """ReLU激活函数.
    """
    def __init__(self, name='relu'):
        super(Relu, self).__init__(name=name)

    def __call__(self, z):
        """
        Arguments:
            z: numpy.ndarray, 输入的张量.

        Returns:
            经过激活后的张量.
        """
        result = np.maximum(0, z)

        return result

    def diff(self, output, a, *args, **kwargs):
        """ReLU函数的微分.

        Arguments:
            output: numpy.ndarray, 前向传播输出的张量.
            a: numpy.ndarray, 输入的张量.

        Notes:
            ReLU函数在大于零区间的导数应该是恒为一, 如果按此计算在实际应用上会随着训练轮数的增加, 最后模型的输出是一个随机概率,
            作者个人认为原因是随着轮数的增加, 大部分神经元都恒为激活态, 成为一个线性操作(缩放实际上相当于不参与计算了).
            在实际应用中使用原值发现可以避免这种想象.
        """
        da = np.asarray(output)
        da[a <= 0] = 0

        return da


class Sigmoid(Activation):
    """Sigmoid激活函数.
    """
    def __init__(self, name='sigmoid'):
        super(Sigmoid, self).__init__(name=name)

    def __call__(self, z):
        """
        Arguments:
            z: numpy.ndarray, 输入的张量.

        Returns:
            经过激活后的张量.
        """
        result = 1 / (1 + np.exp(-z))

        return result

    def diff(self, output, a, *args, **kwargs):
        """Sigmoid的导数(微分).

        Arguments:
            output: numpy.ndarray, 输出张量.
            a: numpy.ndarray, 输入张量.
            args:
                y_true: numpy.ndarray, 真实的标签.

        Returns:
            Sigmoid的导数(微分).

        Notes:
            Sigmoid的导数f' = a * (1 - a),
            但是作为输出层就需要乘上误差.
        """
        y_true = args[0]
        error = y_true - output

        da = a * (1 - a) * error

        return da


class Softmax(Activation):
    """Softmax激活函数.
    """
    def __init__(self, name='softmax'):
        super(Softmax, self).__init__(name=name)

    def __call__(self, z):
        """
        Arguments:
            z: numpy.ndarray, 输入的张量.

        Returns:
            经过激活后的张量.
        """
        z -= np.max(z)  # 为了避免溢出减去最大值
        result = np.exp(z) / np.sum(np.exp(z), axis=1, keepdims=True)

        return result

    def diff(self, output, a, *args, **kwargs):
        """Softmax函数的微分.

        Arguments:
            output: numpy.ndarray, 前向传播输出的张量.
            a: numpy.ndarray, 输入的张量.

        References:
            https://blog.csdn.net/qq_38032064/article/details/90599547?
        """
        da = a - output

        return da


# Instances.
relu = Relu()
sigmoid = Sigmoid()
softmax = Softmax()



import os
import time

import numpy as np
import psutil

from classicML import CLASSICML_LOGGER


def _format_and_display_the_time_spent(start_time=None, end_time=None, time_spent_list=None, repeat=None):
    """格式化并显示运行时间.

    Arguments:
        start_time: float, default=None,
            函数开始运行的时间.
        end_time: float, default=None,
            函数结束的时间.
        time_spent_list: numpy.ndarray, default=None,
            记录多次运行时间的列表.
        repeat: int, default=None,
            重复运行的次数.
    """
    if start_time is not None:
        time_spent = (end_time - start_time) * 1000 * 1000  # 返回时间的单位是s, s -> us.

        if int(time_spent) // 1000 // 1000 > 0:
            CLASSICML_LOGGER.info('耗时 {:.3f} s'.format(time_spent / 1000 / 1000))
        elif int(time_spent) // 1000 > 0:
            CLASSICML_LOGGER.info('耗时 {:.3f} ms'.format(time_spent / 1000))
        else:
            CLASSICML_LOGGER.info('耗时 {:.3f} us'.format(time_spent))
    else:
        time_spent_list = time_spent_list * 1000 * 1000  # 返回时间的单位是s, s -> us.

        average_time_spent = np.mean(time_spent_list)
        std_time = np.std(time_spent_list)
        min_time = np.min(time_spent_list)
        max_time = np.max(time_spent_list)

        if int(max_time) // 1000 // 1000 > 0:
            average_time_spent /= (1000 * 1000)
            std_time /= 1000
            min_time /= (1000 * 1000)
            max_time /= (1000 * 1000)
            unit1, unit2 = 's', 'ms'
        elif int(max_time) // 1000 > 0:
            average_time_spent /= 1000
            min_time /= 1000
            max_time /= 1000
            unit1, unit2 = 'ms', 'us'
        else:
            unit1, unit2 = 'us', 'us'

        CLASSICML_LOGGER.info('平均耗时 {:.3f} %s ± {:.0f} %s, {:.3f} %s, {:.3f} %s; 循环次数 {:d} (mean ± std, max, min)'
                              .format(average_time_spent, std_time, max_time, min_time, repeat)
                              % (unit1, unit2, unit1, unit1))


# TODO(Steve R. Sun, tag:code): 将@timer和@average_timer()合并, 并且可以不使用括号.
def average_timer(repeat=5):
    """程序平均计时装饰器.

    Arguments:
        repeat: int, default=5,
            重复运行的次数.

    Notes:
        - 使用该装饰器统计平均计时会明显降低运行速度,
        请在开发时使用, 避免在训练模型时使用.
    """
    def decorator(function):
        def wrapper(*args, **kwargs):
            return_values = None
            time_spent_list = list()

            for i in range(repeat):
                start_time = time.perf_counter()
                return_values = function(*args, **kwargs)
                end_time = time.perf_counter()

                time_spent_list.append(end_time - start_time)

            _format_and_display_the_time_spent(None, None, np.asarray(time_spent_list), repeat)

            # 函数返回值为最后一次的返回值.
            return return_values

        return wrapper

    return decorator


def memory_monitor(function):
    """内存监视装饰器.

    Notes:
        - 使用该装饰器统计内存信息, 有潜在降低运行速度的可能性.
        并且psutil针对的Python优化手段会导致在CC引擎的速度大幅降低.
    """
    def wrapper(*args, **kwargs):
        return_values = function(*args, **kwargs)

        pid = os.getpid()
        current_process = psutil.Process(pid)
        process_memory = current_process.memory_full_info()
        CLASSICML_LOGGER.info('占用内存 {:.5f} MB'.format(process_memory.uss / 1024 / 1024))
        return return_values

    return wrapper


def timer(function):
    """程序计时装饰器.
    """
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()  # 注意将记录time.sleep()的时间
        return_values = function(*args, **kwargs)
        end_time = time.perf_counter()

        _format_and_display_the_time_spent(start_time, end_time, None, None)

        return return_values

    return wrapper



"""classicML的底层核心操作."""
import numpy as np

from classicML import CLASSICML_LOGGER


def calculate_error(x, y, i, kernel, alphas, non_zero_alphas, b):
    """计算KKT条件的违背值.

    Arguments:
        x: numpy.ndarray, array-like, 特征数据.
        y: numpy.ndarray, array-like, 标签.
        i: int, 第i个样本.
        kernel: classicML.kernel.Kernels 实例, 分类器使用的核函数.
        alphas: numpy.ndarray, 拉格朗日乘子.
        non_zero_alphas: numpy.ndarray, 非零拉格朗日乘子.
        b: float, 偏置项.

    Returns:
        KKT条件的违背值.

    Notes:
        - 该函数提供了非Python后端的实现版本,
          你可以使用其他的版本, 函数的调用方式和接口一致,
          Python版本是没有优化的原始公式版本.
    """
    x_i = x[[i], :]
    y_i = y[i, 0]

    if non_zero_alphas.any():
        # 提取全部合格的标签和对应的拉格朗日乘子.
        valid_x = x[non_zero_alphas]
        valid_y = y[non_zero_alphas]
        valid_alphas = alphas[non_zero_alphas]

        kappa = kernel(valid_x, x_i)

        fx = np.matmul((valid_alphas.reshape(-1, 1) * valid_y).T, kappa.T) + b
    else:
        # 拉格朗日乘子全是零的时候, 每个样本都不会对结果产生影响.
        fx = b

    error = fx - y_i

    return np.squeeze(error)


def clip_alpha(alpha, low, high):
    """修剪拉格朗日乘子.

    Arguments:
        alpha: numpy.ndarray, 拉格朗日乘子.
        low: float, 正则化系数的下界.
        high: float, 正则化系数的上界.

    Returns:
        修剪后的拉格朗日乘子.

    Notes:
        - 该函数提供了非Python后端的实现版本,
          你可以使用其他的版本, 函数的调用方式和接口一致,
          Python版本是没有优化的原始公式版本.
    """
    if alpha > high:
        alpha = high
    elif alpha < low:
        alpha = low

    return alpha


def get_conditional_probability(samples_on_attribute,
                                samples_in_category,
                                num_of_categories,
                                smoothing):
    """获取类条件概率.

    Arguments:
        samples_on_attribute: float, 在某个属性的样本.
        samples_in_category: float, 在某个类别上的样本.
        num_of_categories: int, 类别的数量.
        smoothing: bool, 是否使用平滑.

    Returns:
        类条件概率.

    Notes:
        - 该函数提供了非Python后端的实现版本,
          你可以使用其他的版本, 函数的调用方式和接口一致,
          Python版本是没有优化的原始公式版本.
    """
    if smoothing:
        samples_on_attribute += 1
        samples_in_category += num_of_categories

    probability = samples_on_attribute / samples_in_category

    return probability


def get_dependent_prior_probability(samples_on_attribute_in_category,
                                    number_of_sample,
                                    values_on_attribute,
                                    smoothing):
    """获取有依赖的类先验概率.

    Arguments:
        samples_on_attribute_in_category: int, 类别为c的属性i上取值为xi的样本.
        number_of_sample: int, 样本的总数.
        values_on_attribute: int, 在属性i上的取值数.
        smoothing: bool, 是否使用平滑.

    Returns:
        类先验概率.

    Notes:
        - 该函数提供了非Python后端的实现版本,
          你可以使用其他的版本, 函数的调用方式和接口一致,
          Python版本是没有优化的原始公式版本.
    """
    if smoothing:
        probability = (samples_on_attribute_in_category + 1) / (number_of_sample + 2 * values_on_attribute)  # 执行平滑操作.
    else:
        probability = samples_on_attribute_in_category / number_of_sample

    return probability


def get_prior_probability(number_of_sample, y, smoothing):
    """获取类先验概率.

    Arguments:
        number_of_sample: int, 样本的总数.
        y: numpy.ndarray, 标签.
        smoothing: bool, 是否使用平滑.

    Returns:
        类先验概率.

    Notes:
        - 该函数提供了非Python后端的实现版本,
          你可以使用其他的版本, 函数的调用方式和接口一致,
          Python版本是没有优化的原始公式版本.
    """
    if smoothing:
        p_0 = (len(y[y == 0]) + 1) / (number_of_sample + len(np.unique(y)))  # 执行平滑操作.
    else:
        p_0 = len(y[y == 0]) / number_of_sample

    return p_0, 1 - p_0


def get_probability_density(sample, mean, var):
    """获得概率密度.

    Arguments:
        sample: float, 样本的取值.
        mean: float, 样本在某个属性的上的均值.
        var: float, 样本在某个属性上的方差.

    Returns:
        概率密度.

    Notes:
        - 该函数提供了非Python后端的实现版本,
          你可以使用其他的版本, 函数的调用方式和接口一致,
          Python版本是没有优化的原始公式版本.
    """
    probability = 1 / (np.sqrt(2 * np.pi) * var) * np.exp(-(sample - mean) ** 2 / (2 * var ** 2))

    if probability == 0:
        probability = 1e-36  # probability有可能为零, 导致取对数会有异常, 因此选择一个常小数.

    return probability


def get_w(S_w, mu_0, mu_1):
    """获得投影向量.

    DEPRECATED:
      `ops.get_w` 已经被弃用, 它将在未来的正式版本中被移除, 请使用 `ops.get_w_v2`.

    Arguments:
        S_w: numpy.ndarray, 类内散度矩阵.
        mu_0: numpy.ndarray, 反例的均值向量.
        mu_1: numpy.ndarray, 正例的均值向量.

    Returns:
        投影向量.

    Notes:
        - 该函数提供了非Python后端的实现版本,
          你可以使用其他的版本, 函数的调用方式和接口一致,
          Python版本是没有优化的原始公式版本.
    """
    CLASSICML_LOGGER.warn('`ops.get_w` 已经被弃用, 它将在未来的正式版本中被移除, 请使用 `ops.get_w_v2`.')

    S_w_i = np.linalg.inv(S_w)
    w = np.matmul(S_w_i, (mu_0 - mu_1).T)

    return w.reshape(1, -1)


def get_w_v2(S_w, mu_0, mu_1):
    """获得投影向量.

    Arguments:
        S_w: numpy.ndarray, 类内散度矩阵.
        mu_0: numpy.ndarray, 反例的均值向量.
        mu_1: numpy.ndarray, 正例的均值向量.

    Returns:
        投影向量.

    Notes:
        - 第二版使用奇异值分解来计算类内散度矩阵的逆矩阵.
        - 该函数提供了非Python后端的实现版本,
          你可以使用其他的版本, 函数的调用方式和接口一致,
          Python版本是没有优化的原始公式版本.
    """
    U, Sigma, V_t = np.linalg.svd(S_w)
    S_w_i = V_t.T * np.linalg.inv(np.diag(Sigma)) * U.T

    w = np.matmul(S_w_i, (mu_0 - mu_1).T)

    return w.reshape(1, -1)


def get_within_class_scatter_matrix(X_0, X_1, mu_0, mu_1):
    """获得类内散度矩阵.

    Arguments:
        X_0: numpy.ndarray, 反例集合.
        X_1: numpy.ndarray, 正例集合.
        mu_0: numpy.ndarray, 反例的均值向量.
        mu_1: numpy.ndarray, 正例的均值向量.

    Returns:
        类内散度矩阵.

    Notes:
        - 该函数提供了非Python后端的实现版本,
          你可以使用其他的版本, 函数的调用方式和接口一致,
          Python版本是没有优化的原始公式版本.
    """
    S_0 = np.matmul((X_0 - mu_0).T, (X_0 - mu_0))
    S_1 = np.matmul((X_1 - mu_1).T, (X_1 - mu_1))

    S_w = S_0 + S_1

    return S_w


def select_second_alpha(error, error_cache, non_bound_alphas):
    """选择第二个拉格朗日乘子, SMO采用的是启发式寻找的思想,
    找到目标函数变化量足够大, 即选取变量样本间隔最大.

    Arguments:
        error: float,
            KKT条件的违背值.
        error_cache: numpy.ndarray,
            KKT条件的违背值缓存.
        non_bound_alphas: numpy.ndarray,
            非边界拉格朗日乘子.

    Returns:
        拉格朗日乘子的下标和违背值.

    Notes:
        - 该函数提供了非Python后端的实现版本,
          你可以使用其他的版本, 函数的调用方式和接口一致,
          Python版本是没有优化的原始公式版本.
        - Python和CC存在精度差异, 存在潜在的可能性导致使用同样的数据和随机种子但不同后端的结果不一致,
          不过随着训练的轮数的增加, 这种差异会逐渐消失.
    """
    non_bound_index = non_bound_alphas.nonzero()[0]
    delta_e = np.abs(error - error_cache[non_bound_index])

    index_alpha = non_bound_index[np.argmax(delta_e)]  # 选取间隔最大的下标.
    error_alpha = error_cache[index_alpha]

    return index_alpha, error_alpha


def type_of_target(y):
    """判断输入数据的类型.

    Arguments:
        y: numpy.ndarray,
            待判断类型的数据.

    Returns:
        'binary': 元素只有两个离散值, 类型不限.
        'continuous': 元素都是浮点数, 且不是对应整数的浮点数.
        'multiclass': 元素不只有两个离散值, 类型不限.
        'multilabel': 元素标签不为一, 类型不限.
        'unknown': 类型未知.

    Notes:
        - 该函数提供了非Python后端的实现版本,
          你可以使用其他的版本, 函数的调用方式和接口一致,
          Python版本是没有优化的原始公式版本.
        - Python和CC针对str类型的返回暂不相同.
    """
    if y.dtype == object or np.ndim(y) > 2:
        return 'unknown'

    if y.dtype.kind == 'f' and np.any(y != y.astype(int)):
        return 'continuous'

    if np.ndim(y) == 1:
        if len(np.unique(y)) == 2:
            return 'binary'
        elif len(np.unique(y)) > 2:
            return 'multiclass'
    elif y.shape[1] == 1:
        if len(np.unique(y)) == 2:
            return 'binary'
        elif len(np.unique(y)) > 2:
            return 'multiclass'

    if np.ndim(y) == 2 and len(np.unique(y)) >= 2:
        return 'multilabel'

    return 'unknown'



"""classicML的I/O函数, 用于模型保存等操作."""
import re
from time import time

import h5py
import numpy as np

from classicML import CLASSICML_LOGGER
from classicML import __version__ as cml_version

__version__ = 'backend.io.0.5.1'

min_cml_version = '0.5'
min__version__ = 'backend.io.0.3'


def _parse(model_name, fp):
    """解析描述信息组, 核验文件.

    Arguments:
        model_name: str, 模型的名称.
        fp: h5py._hl.files.File, 文件指针.

    """
    description_gp = fp['description']

    file_cml_version = re.findall('\\d+\\.\\d+(?:\\.\\d+)*', description_gp.attrs['version'])[0]
    file_backend_version = 'backend.io.' + re.findall('\\d+\\.\\d+', description_gp.attrs['version'])[-1]

    if (file_cml_version < min_cml_version) or (file_backend_version < min__version__):
        CLASSICML_LOGGER.error('文件核验失败, 模型版本过低')
        raise ValueError('文件核验失败')
    if description_gp.attrs['model_name'] != model_name:
        CLASSICML_LOGGER.error('文件核验失败, 模型不匹配')
        raise ValueError('文件核验失败')


def initialize_weights_file(filepath, mode, model_name):
    """初始化权重文件, 以创建或者解析模式运行.

    cml.backend.io的HDF5文件标准化协议包括:
      两个信息组description和parameters,
      description用来存放cml兼容性和开发时间等信息; parameters用以保存模型本身的参数.
      parameters分成两个数据集:
        compile保存模型的训练的超参数; weights保存模型的权重信息.
    开发符合标准化协议的自定义模型, 需将固化的参数保存在compile和weights中.

    具体结构如下:
      /
      ｜-- description
          |-- version 版本信息
          |-- model_name 模型名称
          |__ saved_time 时间戳
      ｜__ parameters
          ｜-- compile 训练相关超参数
          ｜__ weights 模型的权重

    Arguments:
        filepath: str, 权重文件加载的路径.
        mode: {'w', 'r'},
            工作模式, 'w'是写入权重文件, 'r'是读取权重文件.
        model_name: str, 模型的名称.

    Returns:
        可操作的文件指针.

    Raises:
        IOError: 初始化失败.
        ValueError: 文件核验失败.
    """
    try:
        fp = h5py.File(filepath, mode)
        if mode == 'w':
            # 创建描述信息组.
            description_gp = fp.create_group(name='description')
            description_gp.attrs['version'] = cml_version + '.' + __version__
            description_gp.attrs['model_name'] = model_name
            description_gp.attrs['saved_time'] = time()

            # 创建参数信息组(包括两个数据集, 分别记录超参数和模型权重).
            parameters_gp = fp.create_group(name='parameters')
            parameters_gp.create_dataset(name='compile', dtype=np.float64)
            parameters_gp.create_dataset(name='weights', dtype=np.float64)
        else:
            # 解析描述信息组.
            _parse(model_name, fp)
            # 提取参数信息组.
            parameters_gp = fp['parameters']
    except IOError:
        CLASSICML_LOGGER.error('模型权重文件初始化失败, 请检查文件')
        raise IOError('初始化失败')

    return parameters_gp



import numpy as np

from classicML import CLASSICML_LOGGER
from classicML.api.models import BaseModel
from classicML.backend import get_within_class_scatter_matrix
from classicML.backend import get_w
from classicML.backend import io


class LinearDiscriminantAnalysis(BaseModel):
    """线性判别分析.

    Attributes:
        w: numpy.ndarray, 投影向量.
        mu_0: numpy.ndarray, 反例的均值.
        mu_1: numpy.ndarray, 正例的均值.
        is_trained: bool, default=False,
            模型训练后将被标记为True.
        is_loaded: bool, default=False,
            如果模型加载了权重将被标记为True.
    """
    def __init__(self):
        super(LinearDiscriminantAnalysis, self).__init__()
        self.w = None
        self.mu_0 = None
        self.mu_1 = None
        self.is_trained = False
        self.is_loaded = False

    def fit(self, x, y, **kwargs):
        """训练模型.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.

        Returns:
            LDA实例.
        """
        x = np.asarray(x)
        y = np.asarray(y)

        X_0 = x[y == 0]
        X_1 = x[y == 1]

        self.mu_0 = np.mean(X_0, axis=0, keepdims=True)
        self.mu_1 = np.mean(X_1, axis=0, keepdims=True)

        # 获得类内散度矩阵
        S_w = get_within_class_scatter_matrix(X_0, X_1, self.mu_0, self.mu_1)
        # 获得投影向量
        self.w = get_w(S_w, self.mu_0, self.mu_1)
        # 标记训练完成
        self.is_trained = True

        return self

    def predict(self, x, **kwargs):
        """模型进行预测.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.

        Returns:
            LDA预测的标签张量.

        Raises:
            ValueError: 模型没有训练的错误.
        """
        if self.is_trained is False and self.is_loaded is False:
            CLASSICML_LOGGER.error('模型没有权重')
            raise ValueError('你必须先进行训练')

        coord = np.dot(x, self.w.T)

        center_0 = np.dot(self.w, self.mu_0.T)
        center_1 = np.dot(self.w, self.mu_1.T)

        y_pred = np.abs(coord - center_0) > np.abs(coord - center_1)
        y_pred = np.squeeze(y_pred)
        y_pred = y_pred.astype(int)

        return y_pred

    def load_weights(self, filepath):
        """加载模型参数.

        Arguments:
            filepath: str, 权重文件加载的路径.

        Raises:
            KeyError: 模型权重加载失败.
        """
        # 初始化权重文件.
        parameters_gp = io.initialize_weights_file(filepath=filepath,
                                                   mode='r',
                                                   model_name='LinearDiscriminantAnalysis')
        # 加载模型参数.
        try:
            weights_ds = parameters_gp['weights']
            self.w = weights_ds.attrs['w']
            self.mu_0 = weights_ds.attrs['mu_0']
            self.mu_1 = weights_ds.attrs['mu_1']
            # 标记加载完成
            self.is_loaded = True
        except KeyError:
            CLASSICML_LOGGER.error('模型权重加载失败, 请检查文件是否损坏')
            raise KeyError('模型权重加载失败')

    def save_weights(self, filepath):
        """将模型权重保存为一个HDF5文件.

        Arguments:
            filepath: str, 权重文件保存的路径.

        Raises:
            TypeError: 模型权重保存失败.
        """
        # 初始化权重文件.
        parameters_gp = io.initialize_weights_file(filepath=filepath,
                                                   mode='w',
                                                   model_name='LinearDiscriminantAnalysis')
        # 保存模型参数.
        try:
            weights_ds = parameters_gp['weights']
            weights_ds.attrs['w'] = self.w
            weights_ds.attrs['mu_0'] = self.mu_0
            weights_ds.attrs['mu_1'] = self.mu_1
        except TypeError:
            CLASSICML_LOGGER.error('模型权重保存失败, 请检查文件是否损坏')
            raise TypeError('模型权重保存失败')



import numpy as np

from classicML import CLASSICML_LOGGER
from classicML.api.models import BaseModel
from classicML.backend import get_kernel
from classicML.backend import get_optimizer
from classicML.backend import io


class SupportVectorClassifier(BaseModel):
    """支持向量分类器.

    Attributes:
        seed: int, default=None,
            随机种子.
        support: numpy.ndarray, default=None,
            分类器的支持向量下标数组.
        support_vector: numpy.ndarray, default=None,
            分类器的支持向量数组.
        support_alpha: numpy.ndarray, default=None,
            分类器的拉格朗日乘子数组.
        support_y: numpy.ndarray, default=None,
            分类器的支持向量对应的标签数组.
        b: float, default=0,
            偏置项.
        C: float, default=1.0,
            软间隔正则化系数.
        kernel: str, classicML.kernel.Kernels 实例, default='rbf'
            分类器使用的核函数.
        gamma: {'auto', 'scale'} or float, default='auto',
            在使用高斯核, sigmoid核或者多项式核时, 核函数系数,
            使用其他核函数时无效.
            - 如果gamma是'auto', gamma = 1 / 训练数据集的特征数.
            - 如果gamma是'scale', gamma = 1 / (训练数据集的特征数 * 训练数据集的方差).
        tol: float, default=1e-3,
            停止训练的最大误差值.
        epochs: int, default=-1,
            最大的训练轮数, 如果是-1则表示需要所有的样本满足条件时,
            才能停止训练, 即没有限制.
        optimizer: classicML.optimizers.Optimizer 实例
            SMO算法.
        is_trained: bool, default=False,
            模型训练后将被标记为True.
        is_loaded: bool, default=False,
            如果模型加载了权重将被标记为True.
    """
    def __init__(self, seed=None):
        """初始化分类器.

        Arguments:
            seed: int, default=None,
                随机种子.
        """
        super(SupportVectorClassifier, self).__init__()
        self.seed = np.random.seed(seed)

        self.support = None
        self.support_vector = None
        self.support_alpha = None
        self.support_y = None
        self.b = 0

        self.C = None
        self.kernel = None
        self.gamma = None
        self.tol = None
        self.epochs = None
        self.optimizer = get_optimizer('SMO')
        self.is_trained = False
        self.is_loaded = False

    def compile(self, C=1.0, kernel='rbf', gamma='auto', tol=1e-3):
        """编译分类器, 配置训练时使用的超参数.

        Arguments:
            C: float, default=1.0,
                软间隔正则化系数.
            kernel: str, classicML.kernel.Kernels 实例, default='rbf',
                分类器使用的核函数.
            gamma: {'auto', 'scale'} or float, default='auto',
                在使用高斯(径向基)核, sigmoid核或者多项式核时, 核函数系数,
                使用其他核函数时无效.
            tol: float, default=1e-3,
                停止训练的最大误差值.
        """
        self.C = C
        self.kernel = kernel
        self.gamma = gamma
        self.tol = tol

    def fit(self, x, y, epochs=1000, **kwargs):
        """训练分类器.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.
            y: numpy.ndarray, array-like, 标签.
            epochs: int, default=1000,
                最大的训练轮数, 如果是-1则表示需要所有的样本满足条件时,
                才能停止训练, 即没有限制.

        Returns:
            SupportVectorClassifier实例.
        """
        x = np.asarray(x)
        y = np.asarray(y)
        if y.ndim == 1:
            y = y.reshape(-1, 1)

        # 数值化gamma
        if self.gamma == 'auto':
            self.gamma = 1.0 / x.shape[1]
        elif self.gamma == 'scale':
            self.gamma = 1.0 / (x.shape[1] * x.var())

        # 获取核函数
        self.kernel = get_kernel(self.kernel, self.gamma)

        # 使用SMO算法优化
        (self.support, self.support_vector, self.support_alpha,
         self.support_y, self.b) = self.optimizer(x, y, self.C, self.kernel, self.tol, epochs)

        # 标记训练完成
        self.is_trained = True

        return self

    def predict(self, x, **kwargs):
        """使用分类器进行预测.

        Arguments:
            x: numpy.ndarray, array-like, 特征数据.

        Returns:
            SupportVectorClassifier预测的张量数组.

        Raises:
            ValueError: 模型没有训练的错误.
        """
        if self.is_trained is False and self.is_loaded is False:
            CLASSICML_LOGGER.error('模型没有训练')
            raise ValueError('你必须先进行训练')

        number_of_sample = x.shape[0]
        y_pred = np.ones((number_of_sample, ), dtype=int)

        for sample in range(number_of_sample):
            kappa = self.kernel(x[sample], self.support_vector)
            fx = np.matmul((self.support_alpha.reshape(-1, 1) * self.support_y).T, kappa.T) + self.b
            if fx < 0:
                y_pred[sample] = -1

        return y_pred

    def load_weights(self, filepath):
        """加载模型参数.

        Arguments:
            filepath: str, 权重文件加载的路径.

        Raises:
            KeyError: 模型权重加载失败.

        Notes:
            模型将不会加载关于优化器的超参数.
        """
        # 初始化权重文件.
        parameters_gp = io.initialize_weights_file(filepath=filepath,
                                                   mode='r',
                                                   model_name='SupportVectorClassifier')
        # 加载模型参数.
        try:
            compile_ds = parameters_gp['compile']
            weights_ds = parameters_gp['weights']

            self.C = compile_ds.attrs['C']
            self.gamma = compile_ds.attrs['gamma']
            self.kernel = get_kernel(compile_ds.attrs['kernel'], self.gamma)
            self.tol = compile_ds.attrs['tol']

            self.support = weights_ds.attrs['support']
            self.support_vector = weights_ds.attrs['support_vector']
            self.support_alpha = weights_ds.attrs['support_alpha']
            self.support_y = weights_ds.attrs['support_y']
            self.b = weights_ds.attrs['b']
            # 标记加载完成
            self.is_loaded = True
        except KeyError:
            CLASSICML_LOGGER.error('模型权重加载失败, 请检查文件是否损坏')
            raise KeyError('模型权重加载失败')

    def save_weights(self, filepath):
        """将模型权重保存为一个HDF5文件.

        Arguments:
            filepath: str, 权重文件保存的路径.

        Raises:
            TypeError: 模型权重保存失败.

        Notes:
            模型将不会保存关于优化器的超参数.
        """
        # 初始化权重文件.
        parameters_gp = io.initialize_weights_file(filepath=filepath,
                                                   mode='w',
                                                   model_name='SupportVectorClassifier')
        # 保存模型参数.
        try:
            compile_ds = parameters_gp['compile']
            weights_ds = parameters_gp['weights']

            compile_ds.attrs['C'] = self.C
            compile_ds.attrs['kernel'] = self.kernel.name
            compile_ds.attrs['gamma'] = self.gamma
            compile_ds.attrs['tol'] = self.tol

            weights_ds.attrs['support'] = self.support
            weights_ds.attrs['support_vector'] = self.support_vector
            weights_ds.attrs['support_alpha'] = self.support_alpha
            weights_ds.attrs['support_y'] = np.asarray(self.support_y, dtype=np.float64)
            weights_ds.attrs['b'] = np.float64(self.b)
        except TypeError:
            CLASSICML_LOGGER.error('模型权重保存失败, 请检查文件是否损坏')
            raise TypeError('模型权重保存失败')



"""classicML中树结构的生成器."""
import pandas as pd

from classicML.backend.training import get_criterion


class _TreeNode(object):
    """树的结点.

    Attributes:
        num_of_leaves: int, default=None,
            叶结点的数量, None表示没有叶结点, 即本身就是叶结点.
        leaf: bool, default=False,
            是否为一个叶结点.
        height: int, default=-1,
            当前结点的高度.
        category: str, default='',
            当前结点的表示的类别.
        feature_name: str, default='',
            当前结点表示划分的类别.
        feature_index: int, default=None,
            当前结点表示划分的类别的下标.
        purity: float, default=None,
            当前结点的纯度.
        continuous: bool, default=False,
            是否是连续属性.
        subtree: dict, default={},
            属性的分支, 即子树.
        dividing_point: float, default=None,
            当前的结点是连续值时, 划分点的值.
    """
    def __init__(self):
        self.num_of_leaves = None
        self.leaf = False
        self.height = -1
        self.category = ''
        self.feature_name = ''
        self.feature_index = None
        self.purity = None
        self.continuous = False
        self.subtree = dict()
        self.dividing_point = None

    def reset(self, category):
        """重置结点.

        Arguments:
            category: str, 当前结点的表示的类别.
        """
        self.num_of_leaves = 1
        self.height = 0
        self.category = category
        self.leaf = True
        self.feature_name = ''
        self.feature_index = None
        self.purity = None
        self.continuous = False
        self.subtree = dict()
        self.dividing_point = None


class TreeGenerator(object):
    """树生成器的基类.

    Attributes:
        name: str, 生成器的名称.
        criterion: {'gain', 'gini', 'entropy'}, default='gain',
            决策树学习的划分方式.
    """
    def __init__(self, name=None, criterion=None):
        """初始化生成器.

        Arguments:
            name: str, 生成器的名称.
            criterion: {'gain', 'gini', 'entropy'}, default='gain',
                决策树学习的划分方式.
        """
        super(TreeGenerator, self).__init__()
        self.name = name
        self.criterion = get_criterion(criterion)

    def __call__(self, *args, **kwargs):
        """功能实现."""
        return self.tree_generate(*args, **kwargs)

    def tree_generate(self, *args, **kwargs):
        """树的生成实现."""
        raise NotImplementedError


class DecisionTreeGenerator(TreeGenerator):
    """决策树生成器.

    Attributes:
        _x: pandas.DataFrame,
            未经训练的原始特征数据.
            保存一份原始的数据必然会占用两倍的内存, 但是划分最优属性后, 应该为所有出现过的属性值进行生成分支(子结点).
    """
    def __init__(self, name='decision_tree_generator', criterion=None):
        super(DecisionTreeGenerator, self).__init__(name=name,
                                                    criterion=criterion)

        self._x = None

    def tree_generate(self, x, y):
        """生成决策树.

        Arguments:
            x: pandas.DataFrame, 特征数据.
            y: pandas.DataFrame, 标签.

        Returns:
            DecisionTreeClassifier实例.
        """
        decision_tree = _TreeNode()
        decision_tree.num_of_leaves = 0

        # 样本属于同一个类别.
        if y.nunique() == 1:
            decision_tree.num_of_leaves += 1
            decision_tree.leaf = True
            decision_tree.height = 0
            decision_tree.category = y.values[0]

            return decision_tree

        # 属性被测试完.
        if x.empty:
            decision_tree.num_of_leaves += 1
            decision_tree.leaf = True
            decision_tree.height = 0
            # TODO(Steve R. Sun, tag:code): 函数value_counts在两类元素值相同时, 不会按照某种顺序排序.
            # 而是随机返回, 因此可能遇到决策树生成不一样的情况. 应该重写此函数.
            decision_tree.category = pd.value_counts(y).index[0]

            return decision_tree

        # 选择最优划分.
        feature_name, list_of_purity = self.criterion.optimal_division(x, y)

        decision_tree.feature_name = feature_name
        decision_tree.feature_index = list(self._x.columns).index(decision_tree.feature_name)
        current_feature_values = self._x.loc[:, decision_tree.feature_name]
        decision_tree.purity = list_of_purity[0]

        if len(list_of_purity) != 1:
            decision_tree.continuous = True
            decision_tree.dividing_point = list_of_purity[1]

            # 使用二分法对连续值进行处理.
            greater_part = '>= {:.3f}'.format(decision_tree.dividing_point)
            less_part = '< {:.3f}'.format(decision_tree.dividing_point)
            decision_tree.subtree[greater_part] = self.tree_generate(
                x.loc[current_feature_values >= decision_tree.dividing_point],
                y.loc[current_feature_values >= decision_tree.dividing_point])
            decision_tree.subtree[less_part] = self.tree_generate(
                x.loc[current_feature_values < decision_tree.dividing_point],
                y.loc[current_feature_values < decision_tree.dividing_point])

            decision_tree.num_of_leaves += (decision_tree.subtree[greater_part].num_of_leaves
                                            + decision_tree.subtree[less_part].num_of_leaves)
            decision_tree.height = (decision_tree.subtree[greater_part].height
                                    + decision_tree.subtree[less_part].height) + 1
        else:
            decision_tree.continuous = False

            feature_values = pd.unique(current_feature_values)
            sub_x = x.drop(decision_tree.feature_name, axis=1)  # 最优属性相当于已被使用掉.

            max_height = -1
            # 为每个属性值生成一个分支.
            for feature_value in feature_values:
                if y[current_feature_values == feature_value].empty is True:
                    decision_tree.subtree[feature_value] = self.tree_generate(
                        sub_x.loc[current_feature_values == feature_value],
                        y)  # 如果为空, 理论上标记是数据集中出现的最多的样本, 但是这样并不好通过递归实现, 于是此处传入父节点的样本标记.
                else:
                    decision_tree.subtree[feature_value] = self.tree_generate(
                        sub_x.loc[current_feature_values == feature_value],
                        y.loc[current_feature_values == feature_value])

                # 更新子树的高度.
                if decision_tree.subtree[feature_value].height > max_height:
                    max_height = decision_tree.subtree[feature_value].height

                decision_tree.num_of_leaves += decision_tree.subtree[feature_value].num_of_leaves

            decision_tree.height = max_height + 1

        return decision_tree


from abc import ABC
from abc import abstractmethod


class BaseModel(ABC):
    """classicML的模型抽象基类, classicML的模型全部继承于此.
    通过继承BaseModel, 至少实现fit和predict方法就可以构建自己的模型.
    """
    def __init__(self, *args, **kwargs):
        super(BaseModel, self).__init__(*args, **kwargs)

    def compile(self, **kwargs):
        """编译模型."""
        pass

    @abstractmethod
    def fit(self, x, y, **kwargs):
        """训练模型.

        Arguments:
            x: array-like, 特征数据
            y: array-like, 标签.

        Raises:
            NotImplementedError: 需要用户自行实现.
        """
        raise NotImplementedError

    @abstractmethod
    def predict(self, x, **kwargs):
        """使用模型进行预测.

        Arguments:
            x: array-like, 特征数据.

        Raises:
            NotImplementedError: 需要用户自行实现.
        """
        raise NotImplementedError

    def load_weights(self, filepath):
        """加载模型参数.

        Arguments:
            filepath: str, 权重文件加载的路径.

        Raises:
            NotImplemented: 需要用户自行实现.
        """
        raise NotImplemented

    def save_weights(self, filepath):
        """将模型权重保存,
         如果您希望您的模型参数收到保护, 可自行实现模型的保存方式;
         如果您希望您的模型开源, 请参照cml.backend.io的协议方式实现参数的保存.

        Arguments:
            filepath: str, 权重文件保存的路径.

        Raises:
            NotImplemented: 需要用户自行实现.
        """
        raise NotImplemented



"""classicML中的数据预处理模块."""
import copy
from abc import ABC
from abc import abstractmethod

import numpy as np
import pandas as pd


class PreProcessor(ABC):
    """预处理器基类,
    预处理器将实现一系列预处理操作, 部分预处理器还有对应的逆操作.

    Attributes:
        name: str, default='preprocessor',
            预处理器名称.

    Raises:
       NotImplementedError: __call__方法需要用户实现.
       NotImplemented: inverse方法需要用户实现.
    """
    def __init__(self, name='preprocessor'):
        """
        Arguments:
            name: str, default='preprocessor',
                预处理器名称.
        """
        self.name = name

    @abstractmethod
    def __call__(self, *args, **kwargs):
        """预处理操作."""
        raise NotImplementedError

    def inverse(self, *args, **kwargs):
        """预处理逆操作."""
        raise NotImplemented


class DummyEncoder(PreProcessor):
    """Dummy编码器.

    Attributes:
        name: str, default='dummy_encoder',
            Dummy编码器名称.
        dtype: str, default='float32',
            编码后的标签的数据类型.
        class_indices: dict,
            类标签和类索引的映射字典.
    """
    def __init__(self, name='dummy_encoder', dtype='float32'):
        """
        Arguments:
            name: str, default='dummy_encoder',
                Dummy编码器名称.
            dtype: str, default='float32',
                编码后的标签的数据类型.
        """
        super(DummyEncoder, self).__init__(name=name)
        self.dtype = dtype

        self.class_indices = dict()

    def __call__(self, labels):
        """进行Dummy编码.

        Arguments:
            labels: array-like, 原始的标签.

        Returns:
            Dummy编码后的标签.
        """
        labels = np.asarray(labels)

        num_classes = np.unique(labels)
        m = len(labels)  # 样本总数.
        n = len(num_classes)  # 类别总数.

        # 构建映射字典.
        for index, value in enumerate(num_classes):
            self.class_indices.update({value: index})

        dummy_label = np.zeros(shape=[m, n], dtype=self.dtype)
        for index, label in enumerate(labels):
            dummy_label[index][self.class_indices[label]] = 1

        return dummy_label


class Imputer(PreProcessor):
    """缺失值填充器,
    连续值将填充均值, 离散值将填充众数.

    Attributes:
        name: str, default='imputer',
            缺失值填充器名称.
    """
    def __init__(self, name='imputer'):
        """
        Arguments:
            name: str, default='imputer',
                缺失值填充器名称.
        """
        super(Imputer, self).__init__(name=name)

    def __call__(self, data):
        """进行缺失值填充.

        Arguments:
            data: array-like, 输入的数据.

        Returns:
            填充后的数据.
        """
        preprocessed_data = copy.deepcopy(data)
        for column in range(data.shape[1]):
            preprocessed_data[:, column] = self._fillna(data[:, column])

        return preprocessed_data

    @staticmethod
    def _fillna(column):
        """填充数据列中的缺失值.

        Arguments:
            column: array-like, 输入的数据列.

        Returns:
            填充后的数据列.
        """
        try:
            new_column = pd.DataFrame(column, dtype='float32')
            new_column.fillna(value=np.mean(new_column.dropna().values),
                              inplace=True)
        except ValueError:
            new_column = pd.DataFrame(column)
            new_column.fillna(value=new_column.value_counts().keys()[0][0],
                              inplace=True)

        return np.squeeze(new_column.values)


class MaxMarginEncoder(PreProcessor):
    """最大化间隔编码器, 对于支持向量机的标签编码需要将编码转换为关于超平面的.

    Attributes:
        name: str, default='max_margin_encoder',
            最大化间隔编码器名称.
        dtype: str, default='float32',
            编码后的标签的数据类型.
        class_indices: dict,
            类标签和类索引的映射字典.
    """
    def __init__(self, name='max_margin_encoder', dtype='float32'):
        """
        Arguments:
            name: str, default='max_margin_encoder',
                最大化间隔编码器名称.
            dtype: str, default='float32',
                编码后的标签的数据类型.
        """
        super(MaxMarginEncoder, self).__init__(name=name)
        self.dtype = dtype

        self.class_indices = dict()

    def __call__(self, labels):
        """进行最大化间隔编码.

        Arguments:
            labels: array-like, 原始的标签.

        Returns:
            最大化间隔编码后的标签, 类标签和类索引的映射字典.
        """
        labels = np.asarray(labels)

        num_classes = np.unique(labels)
        m = len(labels)  # 样本总数.

        # 构建映射字典.
        for index, value in enumerate(num_classes):
            self.class_indices.update({value: (-1 if index == 0 else 1)})

        max_margin_label = np.zeros(m, dtype=self.dtype)
        for i in range(m):
            max_margin_label[i] = self.class_indices[labels[i]]

        return max_margin_label, self.class_indices


class MinMaxScaler(PreProcessor):
    """归一化器.

    Attributes:
        name: str, default='minmax_scalar',
            归一化器的名称.
        dtype: str, default='float32',
            归一化后数据元素的数据类型.
        axis: int, default=-1,
            归一化所沿轴.
        min: float, default=None,
            数据的最小值.
        max: float, default=None,
            数据的最大值.
    """
    def __init__(self, name='minmax_scalar', dtype='float32', axis=-1):
        """
        Arguments:
            name: str, default='minmax_scalar',
                归一化器的名称.
            dtype: str, default='float32',
                归一化后数据元素的数据类型.
            axis: int, default=-1,
                归一化所沿轴.
        """
        super(MinMaxScaler, self).__init__(name=name)
        self.dtype = dtype
        self.axis = axis

        self.min = None
        self.max = None

    def __call__(self, data):
        """进行归一化.

        Arguments:
            data: array-like, 输入的数据.

        Returns:
            归一化后的数据.
        """
        data = np.asarray(data, dtype=self.dtype)
        self.min = np.min(data, axis=self.axis)
        self.max = np.max(data, axis=self.axis)

        preprocessed_data = (data - self.min) / (self.max - self.min)

        return preprocessed_data.astype(self.dtype)

    def inverse(self, preprocessed_data):
        """进行反归一化.

        Arguments:
            preprocessed_data: array-like, 输入的归一化后数据.

        Returns:
            归一化前的数据.
        """
        preprocessed_data = np.asarray(preprocessed_data, dtype=self.dtype)
        data = preprocessed_data * (self.max - self.min) + self.min

        return data.astype(self.dtype)


class OneHotEncoder(PreProcessor):
    """独热编码器.

    Attributes:
        name: str, default='one-hot_encoder',
            独热编码器名称.
        dtype: str, default='float32',
            编码后的标签的数据类型.
        class_indices: dict,
            类标签和类索引的映射字典.
    """
    def __init__(self, name='one-hot_encoder', dtype='float32'):
        """
        Arguments:
            name: str, default='one-hot_encoder',
                独热编码器的名称.
            dtype: str, default='float32',
                编码后的标签的数据类型.
        """
        super(OneHotEncoder, self).__init__(name=name)
        self.dtype = dtype

        self.class_indices = dict()

    def __call__(self, labels):
        """进行独热编码.

        Arguments:
            labels: array-like, 原始的标签.

        Returns:
            独热编码后的标签, 类标签和类索引的映射字典.
        """
        labels = np.asarray(labels)

        num_classes = np.unique(labels)
        m = len(labels)  # 样本总数.
        n = len(num_classes)  # 类别总数.
        _class_indices = dict()

        for index, value in enumerate(num_classes):
            _class_indices.update({value: index})
            # 构建映射字典.
            _label = np.zeros(n, dtype=self.dtype)
            _label[index] = 1
            self.class_indices.update({value: _label.tolist()})

        # 进行逐元素编码操作.
        onehot_label = np.zeros([m, n], dtype=self.dtype)
        for i in np.arange(m):
            j = _class_indices[labels[i]]
            onehot_label[i, j] = 1

        return onehot_label, self.class_indices


class StandardScaler(PreProcessor):
    """标准化器.

    Attributes:
        name: str, default='standard_scalar',
            标准化器的名称.
        dtype: str, default='float32',
            标准化后数据元素的数据类型.
        axis: int, default=-1,
            标准化所沿轴.
        mean: float, default=None,
            数据的均值.
        var: float, default=None,
            数据的方差.
    """
    def __init__(self, name='standard_scalar', dtype='float32', axis=-1):
        """
        Arguments:
            name: str, default='standard_scalar',
                标准化器的名称.
            dtype: str, default='float32',
                标准化后数据元素的数据类型.
            axis: int, default=-1,
                标准化所沿轴.
        """
        super(StandardScaler, self).__init__(name=name)
        self.dtype = dtype
        self.axis = axis

        self.mean = None
        self.var = None

    def __call__(self, data):
        """进行标准化.

        Arguments:
            data: array-like, 输入的数据.

        Returns:
            标准化后的数据.
        """
        data = np.asarray(data, dtype=self.dtype)
        self.mean = np.mean(data, axis=self.axis)
        self.var = np.var(data, axis=self.axis)

        preprocessed_data = (data - self.mean) / self.var

        return preprocessed_data.astype(self.dtype)

    def inverse(self, preprocessed_data):
        """进行反标准化.

        Arguments:
            preprocessed_data: array-like, 输入的标准化后数据.

        Returns:
            标准化前的数据.
        """
        preprocessed_data = np.asarray(preprocessed_data, dtype=self.dtype)
        data = preprocessed_data * self.var + self.mean

        return data.astype(self.dtype)



from pickle import dumps, loads

import numpy as np
import pandas as pd

from classicML import CLASSICML_LOGGER
from classicML.api.models import BaseModel
from classicML.backend import get_pruner
from classicML.backend import io
from classicML.backend import tree


class DecisionTreeClassifier(BaseModel):
    """决策树分类器.

    Attributes:
        attribute_name: list of name, default=None,
            属性的名称.
        tree: classicML.backend.tree._TreeNode实例,
            生成的决策树.
        criterion: {'gain', 'gini', 'entropy'}, default='gain',
            决策树学习的划分方式.
        generator: classicML.backend.tree.DecisionTreeGenerator实例,
            生成决策树的实现算法.
        pruner: classicML.backend.tree.Pruner实例,
            决策树的剪枝器.
        is_trained: bool, default=False,
            模型训练后将被标记为True.
        is_loaded: bool, default=False,
            如果模型加载了权重将被标记为True.
    """
    def __init__(self, attribute_name=None):
        """初始化决策树.

        Arguments:
             attribute_name: list of name, default=None,
                属性的名称.
        """
        super(DecisionTreeClassifier, self).__init__()
        self.attribute_name = attribute_name

        self.tree = None
        self.criterion = None
        self.generator = None
        self.pruner = None

        self.is_trained = False
        self.is_loaded = False

    def compile(self, criterion='gain', pruning=None):
        """编译决策树, 配置训练时使用的超参数.

        Arguments:
            criterion: {'gain', 'gini', 'entropy'}, default='gain',
                决策树学习的划分方式.
            pruning: {None, 'pre', 'post'}, default=None,
                是否对决策树进行剪枝操作.

        Raises:
            AttributeError: 参数错误.
        """
        if criterion not in ('gain', 'gini', 'entropy'):
            CLASSICML_LOGGER.error('请检查你使用的划分方式')
            raise AttributeError('参数错误')
        if pruning not in (None, 'pre', 'post'):
            CLASSICML_LOGGER.error('请检查你输入的操作名称')
            raise AttributeError('参数错误')

        self.criterion = criterion
        self.generator = tree.generators.DecisionTreeGenerator(criterion=self.criterion)
        self.pruner = get_pruner(pruning=pruning)

    def fit(self, x, y, x_validation=None, y_validation=None):
        """训练决策树分类器.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like,
                特征数据.
            y: numpy.ndarray or pandas.DataFrame, array-like,
                标签.
            x_validation: numpy.ndarray or pandas.DataFrame, array-like,
                剪枝使用的验证特征数据.
            y_validation: numpy.ndarray or pandas.DataFrame, array-like,
                剪枝使用的验证标签.

        Returns:
            DecisionTreeClassifier实例.

        Raises:
            AttributeError: 没有验证集.
        """
        if isinstance(x, np.ndarray):
            x = x.tolist()
            if self.attribute_name is None:
                CLASSICML_LOGGER.warn("属性名称缺失, 请使用pandas.DataFrame; 或检查 self.attributes_name")
        if (self.pruner is not None) and (x_validation is None or y_validation is None):
            CLASSICML_LOGGER.error("没有验证集, 无法对决策树进行剪枝")
            raise AttributeError('没有验证集')

        # 为特征数据添加属性信息.
        x = pd.DataFrame(x, columns=self.attribute_name)
        x.reset_index(drop=True, inplace=True)
        self.generator._x = x

        y = pd.Series(y)
        y.reset_index(drop=True, inplace=True)

        # 为验证数据添加属性信息.
        if x_validation is not None:
            x_validation = pd.DataFrame(x_validation, columns=self.attribute_name)
            x_validation.reset_index(drop=True, inplace=True)

            y_validation = pd.Series(y_validation)
            y_validation.reset_index(drop=True, inplace=True)

        # 没有使用权重文件, 则生成决策树分类器.
        if self.is_loaded is False:
            self.tree = self.generator(x, y)

        # 进行剪枝.
        if self.pruner:
            self.tree = self.pruner(x, y, x_validation, y_validation, self.tree)

        # 标记训练完成
        self.is_trained = True

        return self

    def predict(self, x, **kwargs):
        """使用决策树分类器进行预测.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like,
                特征数据.

        Returns:
            DecisionTreeClassifier预测的结果.

        Raises:
            ValueError: 模型没有训练的错误.
        """
        if self.is_trained is False and self.is_loaded is False:
            CLASSICML_LOGGER.error('模型没有训练')
            raise ValueError('你必须先进行训练')

        # 修正数据类型.
        if isinstance(x, list):
            x = np.expand_dims(x, axis=0)
        elif isinstance(x, pd.DataFrame):
            x = x.values
        elif isinstance(x, pd.Series):
            x = np.expand_dims(x.values, axis=0)

        y_pred = list()
        for feature in x:
            y_pred.append(self._predict(feature, self.tree))

        return y_pred

    def _predict(self, x, decision_tree):
        """通过递归决策树预测结果.

        Arguments:
            x: numpy.ndarray, 特征数据.
            decision_tree: classicML.backend.tree._TreeNode, 决策树实例.
        """
        if decision_tree.leaf:
            return decision_tree.category

        if decision_tree.continuous:
            if x[decision_tree.feature_index] >= decision_tree.dividing_point:
                return self._predict(x, decision_tree.subtree['>= {:.3f}'.format(decision_tree.dividing_point)])
            else:
                return self._predict(x, decision_tree.subtree['< {:.3f}'.format(decision_tree.dividing_point)])
        else:
            return self._predict(x, decision_tree.subtree[x[decision_tree.feature_index]])

    def load_weights(self, filepath):
        """加载模型参数.

        Arguments:
            filepath: str, 权重文件加载的路径.

        Raises:
            KeyError: 模型权重加载失败.

        Notes:
            模型将不会加载关于优化器的超参数.
        """
        # 初始化权重文件.
        parameters_gp = io.initialize_weights_file(filepath=filepath,
                                                   mode='r',
                                                   model_name='DecisionTreeClassifier')
        # 加载模型参数.
        try:
            compile_ds = parameters_gp['compile']
            weights_ds = parameters_gp['weights']

            self.criterion = compile_ds.attrs['criterion']
            if self.pruner is not None:
                self.pruner = get_pruner(compile_ds.attrs['pruning'])

            self.tree = loads(weights_ds.attrs['tree'].tobytes())

            # 标记加载完成
            self.is_loaded = True
        except KeyError:
            CLASSICML_LOGGER.error('模型权重加载失败, 请检查文件是否损坏')
            raise KeyError('模型权重加载失败')

    def save_weights(self, filepath):
        """将模型权重保存为一个HDF5文件.

        Arguments:
            filepath: str, 权重文件保存的路径.

        Raises:
            TypeError: 模型权重保存失败.

        References:
            - [如何存储原始的二进制数据](https://docs.h5py.org/en/2.3/strings.html)

        Notes:
            模型将不会保存关于优化器的超参数.
        """
        # 初始化权重文件.
        parameters_gp = io.initialize_weights_file(filepath=filepath,
                                                   mode='w',
                                                   model_name='DecisionTreeClassifier')
        # 保存模型参数.
        try:
            compile_ds = parameters_gp['compile']
            weights_ds = parameters_gp['weights']

            compile_ds.attrs['criterion'] = self.criterion
            if self.pruner is not None:
                compile_ds.attrs['pruning'] = self.pruner.name
            weights_ds.attrs['tree'] = np.void(dumps(self.tree))
        except TypeError:
            CLASSICML_LOGGER.error('模型权重保存失败, 请检查文件是否损坏')
            raise TypeError('模型权重保存失败')



"""classicML的工具类."""
import sys
from time import time

import numpy as np


class ProgressBar(object):
    """训练进度条.

    Attributes:
        ETD: float, 优化器启动的时间戳.
        epochs: int, 训练的轮数.
        loss: str, classicML.losses.Loss 实例,
            模型使用的损失函数.
        metric: str, classicML.metrics.Metric 实例,
            模型使用的评估函数.
    """

    def __init__(self, epochs, loss, metric):
        """
        Arguments:
            epochs: int, 训练的轮数.
            loss: str, classicML.losses.Loss 实例,
                模型使用的损失函数.
            metric: str, classicML.metrics.Metric 实例,
                模型使用的评估函数.
        """
        super(ProgressBar, self).__init__()
        self.ETD = time()
        self.epochs = epochs
        self.loss = loss
        self.metric = metric

    def __call__(self, epoch, current, loss_value, metric_value):
        """函数实现.

        Arguments:
            epoch: int, 当前的训练轮数.
            current: float, 当前的时间戳.
            loss_value: float, 当前的损失值.
            metric_value: float, 当前的评估值.
        """
        self._update_info(epoch, current, loss_value, metric_value)
        self._dynamic_display()

    def _update_info(self, epoch, current, loss_value, metric_value):
        """在终端上显示进度条.

        Arguments:
            epoch: int, 当前的训练轮数.
            current: float, 当前的时间戳.
            loss_value: float, 当前的损失值.
            metric_value: float, 当前的评估值.
        """
        self._draw_bar(epoch)
        self._draw_detail(epoch, current, loss_value, metric_value)

    def _dynamic_display(self):
        """在终端显示.
        """
        sys.stdout.write('\r')
        sys.stdout.write(self.info)
        sys.stdout.flush()

    def _draw_bar(self, epoch):
        """绘制进度条.
        无论总的训练轮数是多少, 显示条的总更新步数是25次.

        Arguments:
            epoch: int, 当前的训练轮数.
        """
        self.info = 'Epoch {}/{} ['.format(epoch, self.epochs)

        if epoch > (self.epochs - self.epochs / 25):
            self.info += '=========================]'
        else:
            # 获取箭头的实时位置
            # arrow = epoch / (epochs / 25)
            arrow = int(np.ceil(epoch / self.epochs * 25))

            self.info += '=' * (arrow - 1)
            self.info += '>'
            self.info += '.' * (25 - arrow)
            self.info += ']'

    def _draw_detail(self, epoch, current, loss_value, metric_value):
        """绘制显示的计算信息.

        Arguments:
            epoch: int, 当前的训练轮数.
            current: float, 当前的时间戳.
            loss_value: float, 当前的损失值.
            metric_value: float, 当前的评估值.
        """
        # 时间信息
        if epoch == 0:
            self.info += ' ETA: 00:00'
        elif epoch > (self.epochs - self.epochs / 25):
            self.total = time() - self.ETD
            self.per_epoch = self.total * 1000 / self.epochs
            self.info += ' {:.0f}s {:.0f}ms/step'.format(self.total, self.per_epoch)
        else:
            ETA = (time() - current) * (self.epochs - epoch)  # 剩余时间
            if ETA < 60:
                self.info += ' ETA: {:.0f}s'.format(ETA)
            else:
                ETA_minutes = int(ETA) // 60
                ETA_seconds = int(ETA) % 60
                self.info += ' ETA: {:02d}:{:02d}'.format(ETA_minutes, ETA_seconds)

        # 数值信息
        self.info += ' - {}: {:.4f} - {}: {:.4f}'.format(self.loss.name, loss_value, self.metric.name, metric_value)

        if epoch == self.epochs:
            self.info += '\n'



import numpy as np
import pandas as pd

from classicML.api.plots import _plt as plt
from classicML.api.plots.utils import _set_bayes_axis
from classicML.api.plots.utils import _bayes_plot_config
from classicML import CLASSICML_LOGGER


def _plot_background(bayes, x):
    """绘制背景分界图.

    Arguments:
        bayes: classicML.models.NB or classicML.models.SPODE,
            朴素贝叶斯分类器或超父独依赖估计器实例.
        x: numpy.ndarray, array-like, 特征数据.
    """
    # 生成向量空间.
    x0_min, x0_max = x.values[:, 0].min() - 0.1, x.values[:, 0].max() + 0.1
    x1_min, x1_max = x.values[:, 1].min() - 0.1, x.values[:, 1].max() + 0.1
    x0_coord = np.linspace(x0_min, x0_max, 300)
    x1_coord = np.linspace(x1_min, x1_max, 300)
    vector_matrix_0, vector_matrix_1 = np.meshgrid(x0_coord, x1_coord)

    # 进行预测并改变形状.
    x_test = pd.DataFrame(np.c_[vector_matrix_0.ravel(), vector_matrix_1.ravel()], columns=x.columns)
    y_pred = bayes.predict(x_test)
    y_pred = np.asarray(y_pred).reshape(vector_matrix_0.shape)

    # 进行绘制.
    plt.pcolormesh(vector_matrix_0, vector_matrix_1, y_pred, cmap='GnBu', alpha=0.75, shading='nearest')


def _plot_scatter(x, y):
    """绘制样本点.

    Arguments:
        x: pandas.core.DataFrame, array-like, 特征数据.
        y: numpy.ndarray, array-like, 标签.

    Returns:
        处理后的数据标签信息.
    """
    # 去除离散值的数据.
    discrete_list = list()
    for i, datatype in enumerate(x.dtypes):
        if datatype != ('float' or 'int'):
            discrete_list.append(x.columns[i])
    x = x.drop(labels=discrete_list, axis=1)
    x = x.iloc[:, :2]  # 由于暂时只能可视化二维空间, 就只保留两个属性.

    negative_label = [y == (0 or False)][0]
    positive_label = [y == (1 or True)][0]

    plt.scatter(x.values[negative_label, 0], x.values[negative_label, 1], c='lightcoral', marker='o', label='反例')
    plt.scatter(x.values[positive_label, 0], x.values[positive_label, 1], c='c', marker='o', label='正例')

    return x.columns


def plot_bayes(bayes, x, y):
    """可视化朴素贝叶斯分类器或超父独依赖估计器的二维示意图.

    Arguments:
        bayes: classicML.models.NB or classicML.models.SPODE,
            朴素贝叶斯分类器或超父独依赖估计器实例.
        x: numpy.ndarray, array-like, 特征数据.
        y: numpy.ndarray, array-like, 标签.

    Raises:
        ValueError: 模型没有训练的错误.
    """
    if bayes.is_trained is False and bayes.is_loaded is False:
        CLASSICML_LOGGER.error('模型没有训练')
        raise ValueError('你必须先进行训练')

    ax = plt.subplot()
    _set_bayes_axis(ax)

    # 转换数据类型.
    x = pd.DataFrame(x, columns=bayes.attribute_name)

    # 绘制背景分界图.
    _plot_background(bayes, x)

    # 绘制样本点.
    x_label, y_label = _plot_scatter(x, y)

    _bayes_plot_config(x_label, y_label)
    plt.show()



"""tree是classicML中关于树操作的模块, 它可以实现树的生成, 剪枝等功能."""
from classicML.backend.python.tree import criteria
from classicML.backend.python.tree import generators
from classicML.backend.python.tree import pruners


import os

from matplotlib import pyplot as _plt

from classicML.api.plots.bayes import plot_bayes
from classicML.api.plots.callbacks import plot_history
from classicML.api.plots.linear_model import plot_linear_discriminant_analysis
from classicML.api.plots.linear_model import plot_linear_discriminant_analysis as plot_lda
from classicML.api.plots.linear_model import plot_logistic_regression
from classicML.api.plots.support_vector_machine import plot_support_vector_classifier
from classicML.api.plots.support_vector_machine import plot_support_vector_classifier as plot_svc
from classicML.api.plots.tree import plot_tree

_plt.rcParams['font.family'] = os.environ.get('CLASSICML_FONT')



"""classicML中的用来组织数据集模块."""
import numpy as np
import pandas as pd

from classicML import CLASSICML_LOGGER
from classicML.backend.python.data.preprocessing import DummyEncoder
from classicML.backend.python.data.preprocessing import Imputer
from classicML.backend.python.data.preprocessing import MaxMarginEncoder
from classicML.backend.python.data.preprocessing import MinMaxScaler
from classicML.backend.python.data.preprocessing import OneHotEncoder
from classicML.backend.python.data.preprocessing import StandardScaler


class Dataset(object):
    """数据集,
    数据集提供了对输入数据的预处理和封装的功能, 使之满足cml模型输入的需要.

    Attributes:
        dataset_type: {'train', 'validation', 'test'}, default='train',
            数据集的类型, 如果声明为测试集, 将不会生成对应的标签.
        label_mode: {'one-hot', 'max-margin'}, default=None,
            标签的编码格式.
        fillna: bool, default=True,
            是否填充缺失值.
        digitization: bool, default=False,
            是否使用数值化, 将离散标签转化成数值.
        normalization: bool, default=False,
            是否使用归一化.
        standardization: bool, default=False,
            是否使用标准化.
        name: str, default=None,
            数据集的名称.
        x: numpy.ndarray,
            处理后的特征数据.
        y: numpy.ndarray,
            处理后的标签.
        class_indices: dict,
            类标签和类索引的映射字典.
    """
    def __init__(self,
                 dataset_type='train',
                 label_mode=None,
                 fillna=True,
                 digitization=False,
                 normalization=False,
                 standardization=False,
                 name=None):
        """初始化数据集.

        Arguments:
            dataset_type: {'train', 'validation', 'test'}, default='train',
                数据集的类型, 如果声明为测试集, 将不会生成对应的标签.
            label_mode: {'one-hot', 'max-margin'}, default=None,
                标签的编码格式.
            fillna: bool, default=True,
                是否填充缺失值.
            digitization: bool, default=False,
                是否使用数值化, 将离散标签转化成数值.
            normalization: bool, default=False,
                是否使用归一化.
            standardization: bool, default=False,
                是否使用标准化.
            name: str, default=None,
                数据集的名称.
        """
        super(Dataset, self).__init__()
        self.dataset_type = dataset_type.lower()
        self.label_mode = label_mode
        self.fillna = fillna
        self.digitization = digitization
        self.normalization = normalization
        self.standardization = standardization
        self.name = name

        self.x = None
        self.y = None
        self.class_indices = dict()

    def from_dataframe(self, dataframe):
        """从DataFrame中加载数据集.

        Arguments:
            dataframe: pandas.DataFrame, 原始的数据.

        Returns:
            经过预处理的特征数据和标签.
        """
        # 预处理特征数据.
        self._preprocessing_features(dataframe.iloc[:, :-1].values)
        # 预处理标签.
        if self.dataset_type != 'test':
            if len(np.unique(dataframe.iloc[:, -1].values)) == 2:
                self._preprocessing_binary_labels(dataframe.iloc[:, -1].values)
            else:
                self._preprocessing_categorical_labels(dataframe.iloc[:, -1].values)
            # 编码标签.
            self._encoder_label()

        return self.x, self.y

    def from_csv(self, filepath, sep=','):
        """从CSV文件中加载数据集,
         也可以从其他的结构化文本读入数据, 例如TSV等.

        Arguments:
            filepath: str, CSV文件的路径.
            sep: str, default=',',
                使用的文本分隔符.

        Returns:
            经过预处理的特征数据和标签.
        """
        data = pd.read_csv(filepath_or_buffer=filepath,
                           sep=sep,
                           index_col=0,
                           header=0)
        self.from_dataframe(data)

        return self.x, self.y

    def from_tensor_slices(self, x, y=None):
        """从张量流加载数据集.

        Arguments:
            x: array-like,
                处理后的特征数据.
            y: array-like, default=None,
                处理后的标签.

        Returns:
            经过预处理的特征数据和标签.
        """
        if y is None and self.dataset_type == 'train':
            CLASSICML_LOGGER.warn('请检查您的数据集, 训练集似乎应该有标签数据.')

        x = np.asarray(x)
        if y is not None:
            y = np.asarray(y)
            if len(y.shape) == 1:
                y = y.reshape(-1, 1)

        # 预处理特征数据.
        self._preprocessing_features(x)

        # 预处理标签.
        if self.dataset_type != 'test':
            if y.shape[1] > 2:
                self._preprocessing_categorical_labels(y)
            else:
                self._preprocessing_binary_labels(y)
            self._encoder_label()

        return self.x, self.y

    def _preprocessing_features(self, features):
        """预处理特征值.

        Arguments:
            features: numpy.ndarray, 特征数据.
        """
        # 处理缺失值.
        if self.fillna:
            features = Imputer()(features)

        _del_list = list()
        for column in range(features.shape[1]):
            try:
                # 连续值处理.
                _current_feature = np.asarray(features[:, column], dtype='float32')
                if self.standardization:
                    features[:, column] = StandardScaler(axis=0)(_current_feature)
                if self.normalization:
                    features[:, column] = MinMaxScaler()(_current_feature)
            except ValueError:
                # 离散值处理.
                if self.digitization:
                    features = np.c_[features, DummyEncoder()(features[:, column])]
                    _del_list.append(column)

        # 删除被数值化的列.
        if len(_del_list) > 0:
            features = np.delete(features, _del_list, axis=1)

        self.x = features

    def _preprocessing_binary_labels(self, labels):
        """预处理二值标签, 将标签转换为数值化的稀疏向量.

        Arguments:
            labels: numpy.ndarray, 原始的标签.
        """
        _raw_labels = np.unique(labels)

        _positive_key_words = ('是', 'y', 'yes', 'Yes', 'YES', 'Y')
        _negative_key_words = ('否', 'n', 'no', 'No', 'NO', 'N')

        if type(labels[0]) not in (int, np.ndarray):
            for key_word in _positive_key_words:
                labels[labels == key_word] = 1
            for key_word in _negative_key_words:
                labels[labels == key_word] = 0

        self.y = labels.astype(int)
        self.class_indices = {_raw_labels[0]: 0, _raw_labels[1]: 1}

    def _preprocessing_categorical_labels(self, labels):
        """预处理多分类标签, 将标签转换为数值化的稀疏向量.

            Arguments:
                labels: numpy.ndarray, 原始的标签.
        """
        _raw_labels = np.unique(labels)

        if type(labels[0]) is not int:
            for index, raw_label in enumerate(_raw_labels):
                labels[labels == raw_label] = index
                self.class_indices.update({raw_label: index})

        self.y = labels.astype(int)

    def _encoder_label(self):
        """编码标签."""
        if self.label_mode == 'one-hot':
            self.y, _class_indices = OneHotEncoder()(self.y)
            self._set_real_class_indices(self.class_indices, _class_indices)
        elif self.label_mode == 'max-margin':
            self.y, _class_indices = MaxMarginEncoder()(self.y)
            self._set_real_class_indices(self.class_indices, _class_indices)

    def _set_real_class_indices(self, preprocessed_class_indices, encoder_class_indices):
        """设置真实的类标签和类索引的映射字典.

        Arguments:
            preprocessed_class_indices: dict, 预处理后的类标签和类索引的映射字典.
            encoder_class_indices: dict, 编码器的类标签和类索引的映射字典.
        """
        class_indices = dict()
        for key, value in preprocessed_class_indices.items():
            class_indices.update({key: encoder_class_indices[value]})

        self.class_indices = class_indices



"""classicML中决策树的划分标准."""
import os
import numpy as np
import pandas as pd

if os.environ['CLASSICML_ENGINE'] == 'CC':
    from classicML.backend.cc.ops import cc_type_of_target as type_of_target
else:
    from classicML.backend.python.ops import type_of_target


class Criterion(object):
    """划分标准基类."""
    def __init__(self, name=None):
        """
        Arguments:
            name: str, default=None,
                划分标准的名称.
        """
        super(Criterion, self).__init__()
        self.name = name

    def __call__(self, D):
        """划分标准算法实现.

        Arguments:
            D: pandas.Series,
                需要计算的数据集.
        """
        raise NotImplementedError

    def get_value(self, *args, **kwargs):
        """计算划分标准的值.

        Arguments:
            args:
                D: pandas.Series, 需要计算的数据集.
                y: pandas.DataFrame, 对应的标签.
                continuous: bool, 是否是连续属性.
        """
        D, y, continuous = args[0], args[1], args[2]

        num_of_labels = y.shape[0]
        unique_features = pd.unique(D)

        if continuous:
            unique_features.sort()
            criterion_value = float('inf')
            dividing_point = None

            # 构建连续值候选划分点集.
            T_set = [
                ((unique_features[i] + unique_features[i + 1]) / 2) for i in range(len(unique_features) - 1)
            ]

            for t in T_set:
                Dv_up = y[D >= t]
                Dv_down = y[D < t]
                current_criterion_value = (Dv_up.shape[0] / num_of_labels * self.__call__(Dv_up)
                                           + Dv_down.shape[0] / num_of_labels * self.__call__(Dv_down))

                if criterion_value > current_criterion_value:
                    criterion_value = current_criterion_value
                    dividing_point = t

            return [criterion_value, dividing_point]
        else:
            criterion_value = 0
            for feature in unique_features:
                Dv = y[feature == D]
                criterion_value += Dv.shape[0] / num_of_labels * self.__call__(Dv)

            return [criterion_value]

    def optimal_division(self, x, y):
        """最优的划分属性.

        Arguments:
            x: pandas.DataFrame, 特征数据.
            y: pandas.DataFrame, 标签.
        """
        attribute_name = x.columns

        # 初始化为list形式是为了保证比较的时候可以都进行数值的比较,
        # 如果遇到连续值更新的时候会返回值和划分点.
        best_criterion_value = [float('inf')]
        best_attribute = None

        for attribute in attribute_name:
            # 需要先判断是连续值还是离散值.
            continuous = (type_of_target(np.asarray(x[attribute])) == 'continuous')
            # 计算基尼指数.
            criterion_value = self.get_value(x[attribute], y, continuous)
            if best_criterion_value[0] > criterion_value[0]:
                best_criterion_value = criterion_value
                best_attribute = attribute

        return best_attribute, best_criterion_value


class Entropy(Criterion):
    """信息熵."""
    def __init__(self, name='entropy'):
        super(Entropy, self).__init__(name=name)

    def __call__(self, D):
        """计算信息熵.

        Arguments:
            D: pandas.Series,
                需要计算的数据集.
        """
        pk = pd.value_counts(D) / D.shape[0]
        value = np.sum(-pk * np.log2(pk))

        return value


class Gain(Entropy):
    """信息增益."""
    def __init__(self, name='gain'):
        super(Gain, self).__init__(name=name)

    def get_value(self, D, y, D_entropy, continuous):
        """计算信息增益.

        Arguments:
            D: pandas.Series, 需要计算的数据集.
            y: pandas.DataFrame, 对应的标签.
            D_entropy: float, 整个数据集的信息熵.
            continuous: bool, 是否是连续属性.
        """
        if continuous:
            entropy, dividing_point = super(Gain, self).get_value(D, y, continuous)
            # 计算信息增益.
            gain = D_entropy - entropy

            return [gain, dividing_point]
        else:
            entropy = super(Gain, self).get_value(D, y, continuous)
            gain = D_entropy - entropy

            return [gain]

    def optimal_division(self, x, y):
        """最优的划分属性.

        Arguments:
            x: pandas.DataFrame, 特征数据.
            y: pandas.DataFrame, 标签.
        """
        attribute_name = x.columns

        # 初始化为list形式是为了保证比较的时候可以都进行数值的比较,
        # 如果遇到连续值更新的时候会返回值和划分点.
        best_gain_value = [float('-inf')]
        best_attribute = None
        # 计算整个数据集的信息熵.
        D_entropy = self.__call__(y)

        for attribute in attribute_name:
            # 需要先判断是连续值还是离散值.
            continuous = (type_of_target(np.asarray(x[attribute])) == 'continuous')
            # 计算信息增益.
            gain_value = self.get_value(x[attribute], y, D_entropy, continuous)
            if best_gain_value[0] < gain_value[0]:
                best_gain_value = gain_value
                best_attribute = attribute

        return best_attribute, best_gain_value


class Gini(Criterion):
    """基尼指数."""
    def __init__(self, name='gini'):
        super(Gini, self).__init__(name=name)

    def __call__(self, D):
        """计算基尼指数.

        Arguments:
            D: pandas.Series,
                需要计算的数据集.
        """
        pk = pd.value_counts(D) / D.shape[0]
        value = 1 - np.sum(pk ** 2)

        return value


from pickle import loads, dumps

import numpy as np
import pandas as pd

from classicML import CLASSICML_LOGGER
from classicML.api.models import BaseModel
from classicML.backend import type_of_target
from classicML.backend import get_conditional_probability
from classicML.backend import get_prior_probability
from classicML.backend import get_probability_density
from classicML.backend import io


class NaiveBayesClassifier(BaseModel):
    """朴素贝叶斯分类器.

    Attributes:
        attribute_name: list of name, default=None,
            属性的名称.
        smoothing: bool, default=True,
            是否使用平滑.
        p_0: float, 反例的类先验概率.
        p_1: float, 正例的类先验概率.
        pi_0: dict, 反例的类条件概率(概率密度).
        pi_1: dict, 正例的类条件概率(概率密度).
        is_trained: bool, default=False,
            模型训练后将被标记为True.
        is_loaded: bool, default=False,
            如果模型加载了权重将被标记为True.
    """
    def __init__(self, attribute_name=None):
        """初始化朴素贝叶斯分类器.

        Arguments:
            attribute_name: list of name, default=None,
                属性的名称.
        """
        super(NaiveBayesClassifier, self).__init__()
        self.attribute_name = attribute_name

        self.p_0 = None
        self.p_1 = None
        self.pi_0 = dict()
        self.pi_1 = dict()
        self.smoothing = None
        self.is_trained = False
        self.is_loaded = False

    def compile(self, smoothing=True):
        """编译朴素贝叶斯分类器.

        Arguments:
            smoothing: bool, default=True,
                是否使用平滑, 这里的实现是拉普拉斯修正.
        """
        self.smoothing = smoothing

    def fit(self, x, y, **kwargs):
        """训练朴素贝叶斯分类器.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like, 特征数据
            y: numpy.ndarray or pandas.DataFrame, array-like, 标签.

        Returns:
            NaiveBayesClassifier实例.
        """
        if isinstance(x, np.ndarray) and self.attribute_name is None:
            CLASSICML_LOGGER.warn("属性名称缺失, 请使用pandas.DataFrame; 或检查 self.attributes_name")

        # 为特征数据添加属性信息.
        x = pd.DataFrame(x, columns=self.attribute_name)
        x.reset_index(drop=True, inplace=True)
        y = pd.Series(y)
        y.reset_index(drop=True, inplace=True)

        # 获取反正例的样本总数.
        negative_samples = x[y == 0]
        positive_samples = x[y == 1]
        num_of_negative_samples = len(negative_samples)
        num_of_positive_samples = len(positive_samples)

        # 获取类先验概率P(c).
        self.p_0, self.p_1 = get_prior_probability(len(x.values), y.values, self.smoothing)

        number_of_samples, number_of_attributes = x.shape
        # 获取每个属性类条件概率P(x_i|c)或类概率密度p(x_i|c)所需的信息.
        for attribute in range(number_of_attributes):
            xi = x.iloc[:, attribute]
            continuous = (type_of_target(xi.values) == 'continuous')

            xi0 = negative_samples.iloc[:, attribute]
            xi1 = positive_samples.iloc[:, attribute]
            if continuous:
                # 连续值概率密度函数信息.
                xi0_mean = np.mean(xi0)
                xi1_mean = np.mean(xi1)

                xi0_var = np.var(xi0)
                xi1_var = np.var(xi1)

                self.pi_0.update({x.columns[attribute]: {
                                  'continuous': continuous,
                                  'values': [xi0_mean, xi0_var]}})  # values存放了均值和方差.
                self.pi_1.update({x.columns[attribute]: {
                                  'continuous': continuous,
                                  'values': [xi1_mean, xi1_var]}})
            else:
                # 离散值计算条件概率信息.
                unique_value = xi.unique()
                num_of_unique_value = len(unique_value)

                xi0_value_count = pd.DataFrame(np.zeros((1, num_of_unique_value)), columns=unique_value)
                xi1_value_count = pd.DataFrame(np.zeros((1, num_of_unique_value)), columns=unique_value)

                for key in pd.value_counts(xi0).keys():
                    xi0_value_count[key] += pd.value_counts(xi0)[key]
                for key in pd.value_counts(xi1).keys():
                    xi1_value_count[key] += pd.value_counts(xi1)[key]

                # 统计不同属性值的样本总数.
                D_c_xi0 = dict()
                D_c_xi1 = dict()
                for index, name in enumerate(pd.value_counts(xi).keys()):
                    D_c_xi0.update({name: np.squeeze(xi0_value_count.values)[index]})
                    D_c_xi1.update({name: np.squeeze(xi1_value_count.values)[index]})

                self.pi_0.update({x.columns[attribute]: {
                    'continuous': continuous,
                    # values存放了每个样本的数量, 在某个类别上的样本总数, 类别的数量.
                    'values': [D_c_xi0, num_of_negative_samples, num_of_unique_value],
                    'smoothing': self.smoothing}})
                self.pi_1.update({x.columns[attribute]: {
                    'continuous': continuous,
                    'values': [D_c_xi1, num_of_positive_samples, num_of_unique_value],
                    'smoothing': self.smoothing}})

        self.is_trained = True

        return self

    def predict(self, x, probability=False, **kwargs):
        """使用朴素贝叶斯分类器进行预测.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like, 特征数据.
            probability: bool, default=False,
                是否使用归一化的概率形式.

        Returns:
            NaiveBayesClassifier的预测结果,
            不使用概率形式将返回0或1的标签数组, 使用将返回反正例概率的数组.

        Raises:
            ValueError: 模型没有训练的错误.
        """
        if self.is_trained is False and self.is_loaded is False:
            CLASSICML_LOGGER.error('模型没有训练')
            raise ValueError('你必须先进行训练')

        # 为特征数据添加属性信息.
        x = pd.DataFrame(x, columns=self.attribute_name)
        x.reset_index(drop=True, inplace=True)

        # 避免下溢进行对数处理.
        p_0 = np.log(self.p_0)
        p_1 = np.log(self.p_1)

        y_pred = list()
        if len(x.shape) == 1:
            y_pred.append(self._predict(x, p_0, p_1, probability))
        else:
            for i in range(x.shape[0]):
                x_test = x.iloc[i, :]
                y_pred.append(self._predict(x_test, p_0, p_1, probability))

        return y_pred

    def _predict(self, x, p_0, p_1, probability):
        """通过朴素贝叶斯分类器预测单个样本.

        Arguments:
            x: numpy.ndarray or pandas.DataFrame, array-like, 特征数据.
            p_0: float, 反例的类先验概率.
            p_1: float, 正例的类先验概率.
            probability: bool, default=False,
                是否使用归一化的概率形式.

        Returns:
            返回预测的结果.
        """
        for index, attribute in enumerate(self.attribute_name):
            pxi_0 = self.pi_0[attribute]
            pxi_1 = self.pi_1[attribute]

            # 正例连续反例必然连续.
            if pxi_0['continuous']:
                mean0, var0 = pxi_0['values']
                mean1, var1 = pxi_1['values']

                p_0 += np.log(get_probability_density(x[index], mean0, var0))
                p_1 += np.log(get_probability_density(x[index], mean1, var1))
            else:
                D_c_x0, D_c0, N0 = pxi_0['values']
                D_c_x1, D_c1, N1 = pxi_1['values']

                p_0 += np.log(get_conditional_probability(D_c_x0[x[index]], D_c0, N0, self.smoothing))
                p_1 += np.log(get_conditional_probability(D_c_x1[x[index]], D_c1, N1, self.smoothing))

        if probability:
            return [p_0 / (p_0 + p_1), p_1 / (p_0 + p_1)]
        else:
            if p_0 > p_1:
                return 0
            else:
                return 1

    def load_weights(self, filepath):
        """加载模型参数.

        Arguments:
            filepath: str, 权重文件加载的路径.

        Raises:
            KeyError: 模型权重加载失败.

        Notes:
            模型将不会加载关于优化器的超参数.
        """
        # 初始化权重文件.
        parameters_gp = io.initialize_weights_file(filepath=filepath,
                                                   mode='r',
                                                   model_name='NaiveBayesClassifier')
        # 加载模型参数.
        try:
            compile_ds = parameters_gp['compile']
            weights_ds = parameters_gp['weights']

            self.smoothing = compile_ds.attrs['smoothing']
            self.p_0 = weights_ds.attrs['p_0']
            self.p_1 = weights_ds.attrs['p_1']
            self.pi_0 = loads(weights_ds.attrs['pi_0'].tobytes())
            self.pi_1 = loads(weights_ds.attrs['pi_1'].tobytes())

            # 标记加载完成
            self.is_loaded = True
        except KeyError:
            CLASSICML_LOGGER.error('模型权重加载失败, 请检查文件是否损坏')
            raise KeyError('模型权重加载失败')

    def save_weights(self, filepath):
        """将模型权重保存为一个HDF5文件.

        Arguments:
            filepath: str, 权重文件保存的路径.

        Raises:
            TypeError: 模型权重保存失败.

        Notes:
            模型将不会保存关于优化器的超参数.
        """
        # 初始化权重文件.
        parameters_gp = io.initialize_weights_file(filepath=filepath,
                                                   mode='w',
                                                   model_name='NaiveBayesClassifier')
        # 保存模型参数.
        try:
            compile_ds = parameters_gp['compile']
            weights_ds = parameters_gp['weights']

            compile_ds.attrs['smoothing'] = self.smoothing
            weights_ds.attrs['p_0'] = self.p_0
            weights_ds.attrs['p_1'] = self.p_1
            weights_ds.attrs['pi_0'] = np.void(dumps(self.pi_0))
            weights_ds.attrs['pi_1'] = np.void(dumps(self.pi_1))
        except TypeError:
            CLASSICML_LOGGER.error('模型权重保存失败, 请检查文件是否损坏')
            raise TypeError('模型权重保存失败')
